import os
from gql import gql

from alectio_sdk.api.base_attribute import BaseAttribute
from alectio_sdk.tools.utils import extract_id
from alectio_sdk.tools.parser import ParseStrategyYaml
from alectio_sdk.tools.mutations import START_EXPERIMENT_MUTATION, UPLOAD_QUERY_STRATEGY_MUTATION
from alectio_sdk.tools.fragments import USER_PAID_QUERY_FRAGMENT


class Experiment(BaseAttribute):

    def __init__(self, client, attr, user_id, id):
        self._client = client
        self._attr = attr # experiment attributes
        self._name = attr['name']
        self._user_id = user_id
        self._id = id
        self._project_id = ""
        super().__init__(self._attr, self._id)
        self.set_project_id() # set experiment id based on partitition or sort key


    # TODO: start_auto ? if its auto al then you do not need to upload anything.

    def start(self):
        """
        start an experiment from the sdk
        """
        # parse yaml
        print("starting an experiment")
        query = gql(START_EXPERIMENT_MUTATION)
        params = {
            "userId": self._user_id,
            "projectId": self._project_id,
            "experimentId": self._id
        }

        res = self._client.execute(query, params)
        return

    def upload_query_strategy(self, strategy_path):
        """
        upload a qs before an experiment is run.
        if the user has created an experiment for manual al, then
        the user will have to upload a qs at every loop.
        if the user has created an experiment for regular al, then the
        user will have to upload a qs at the beginning.
        if the user is using auto al, then the user does not have to upload
        a qs.
        :params: a yaml file containing a strategy to use for the experiment.
        """

        # TODO: add assertion
        # if strategy_path is None or not os.path.exists(strategy_path):
        # # need to upload, the qs list + mode
        #     raise "Path to query strategies not found"


        #TODO: tell the user somewhere that if they have multiple of the same QS in expert mode they need to
        #distinguish the new QS with an _

        query = gql(USER_PAID_QUERY_FRAGMENT)
        params = {
            "id": self._user_id,
        }
        res = self._client.execute(query, params)

        #Check if the user is a free or paid user, before we allow them to upload a querying strategy.

        if (not res["user"][0]["isPaid"]):
            print("You must be a paid user to upload a YAML file containing your strategy")
            return


        # # parse yaml and check for any issues wihtin the file
        strategies = ParseStrategyYaml(strategy_path)

        experiment_mode = strategies.experiment_mode
        query_strategy_list = strategies.qs_list
        experiment_type = strategies.experiment_type

        print("this is my qs list")
        print(query_strategy_list)
        print("##################")
        print(experiment_type)
        # convert all n_rec cases to camel case for grqphql
        params = {
            "queryStratData": query_strategy_list,
            "projectId": self._project_id,
            "experimentId": self._id,
            "type": experiment_type,
            "mode": experiment_mode
        }

        query = gql(UPLOAD_QUERY_STRATEGY_MUTATION)
        # make sure the backend airlfow gets triggered
        message = self._client.execute(query, params)
        print(message)
        # send the information to the backend to process
        return


    def set_project_id(self):
        """
        return the project associated with the experiment id
        depending if the experiment comes from the get_collection query
        or a get_single query the experiment id can be fromn the
        sort key or primary key
        """
        pk = extract_id(self._attr['pk'])
        sk = extract_id(self._attr['sk'])
        self._project_id = pk
        if pk == self._id:
            self._project_id = sk
        return

    def to_camel_case(self, snake_str):
        components = snake_str.split('_')
        # We capitalize the first letter of each component except the first one
        # with the 'title' method and join them together.
        return components[0] + ''.join(x.title() for x in components[1:])


    def __repr__(self):
        return "<Experiment {}>".format(self._name)







"""
Base for all incomming objects.
"""

# there may need to be an iterable for incomming objects
class BaseAttribute:
    def __init__(self, objects, id):
        self._objects = objects
        self._id = id # id associated with the base class

    @property
    def id(self):
        """
        id associated with the base class
        """
        return self._id 

    @property
    def objects(self):
        """
        objects associated with the class
        """
        for key, value in self._objects.items():
            print(key, ' : ', value)
            
        return self._objects

    #TODO: convert camel case to snakecase, and use these attributes as the parent class + gettr methods




# -*- coding: utf-8 -*-
"""Handle Manager."""

from __future__ import print_function

import json
import logging
import numbers
import os

import six
from alectio_helpers.internal import meta, sample, stats
from alectio_helpers.internal import tb_watcher
from alectio_helpers.lib import proto_util
from alectio_helpers.proto import alectio_helpers_internal_pb2


logger = logging.getLogger(__name__)


class HandleManager(object):
    def __init__(
        self, settings, record_q, result_q, stopped, sender_q, writer_q, interface,
    ):
        self._settings = settings
        self._record_q = record_q
        self._result_q = result_q
        self._stopped = stopped
        self._sender_q = sender_q
        self._writer_q = writer_q
        self._interface = interface

        self._tb_watcher = None
        self._system_stats = None

        # keep track of config and summary from key/val updates
        # self._consolidated_config = dict()
        self._consolidated_summary = dict()
        self._sampled_history = dict()

    def handle(self, record):
        record_type = record.WhichOneof("record_type")
        assert record_type
        handler_str = "handle_" + record_type
        handler = getattr(self, handler_str, None)
        assert handler, "unknown handle: {}".format(handler_str)
        handler(record)

    def handle_request(self, record):
        request_type = record.request.WhichOneof("request_type")
        assert request_type
        handler_str = "handle_request_" + request_type
        handler = getattr(self, handler_str, None)
        logger.debug("handle_request: {}".format(request_type))
        assert handler, "unknown handle: {}".format(handler_str)
        handler(record)

    def _dispatch_record(self, record, always_send=False):
        if not self._settings._offline or always_send:
            self._sender_q.put(record)
        if not record.control.local:
            self._writer_q.put(record)

    def handle_request_defer(self, record):
        defer = record.request.defer
        state = defer.state

        logger.info("handle defer: {}".format(state))
        # only handle flush tb (sender handles the rest)
        if state == defer.FLUSH_STATS:
            if self._system_stats:
                # TODO(jhr): this could block so we dont really want to call shutdown
                # from handler thread
                self._system_stats.shutdown()
        elif state == defer.FLUSH_TB:
            if self._tb_watcher:
                # shutdown tensorboard workers so we get all metrics flushed
                self._tb_watcher.finish()
                self._tb_watcher = None
        elif state == defer.FLUSH_SUM:
            self._save_summary(self._consolidated_summary, flush=True)

        # defer is used to drive the sender finish state machine
        self._dispatch_record(record, always_send=True)

    def handle_request_login(self, record):
        self._dispatch_record(record)

    def handle_run(self, record):
        self._dispatch_record(record)

    def handle_stats(self, record):
        self._dispatch_record(record)

    def handle_config(self, record):
        self._dispatch_record(record)

    def handle_output(self, record):
        self._dispatch_record(record)

    def handle_files(self, record):
        self._dispatch_record(record)

    def handle_artifact(self, record):
        self._dispatch_record(record)

    def _save_summary(self, summary_dict, flush=False):
        summary = alectio_helpers_internal_pb2.SummaryRecord()
        for k, v in six.iteritems(summary_dict):
            update = summary.update.add()
            update.key = k
            update.value_json = json.dumps(v)
        record = alectio_helpers_internal_pb2.Record(summary=summary)
        if flush:
            self._dispatch_record(record)
        elif not self._settings._offline:
            self._sender_q.put(record)

    def _save_history(self, record):
        for item in record.history.item:
            # TODO(jhr) save nested keys?
            k = item.key
            v = json.loads(item.value_json)
            if isinstance(v, numbers.Real):
                self._sampled_history.setdefault(k, sample.UniformSampleAccumulator())
                self._sampled_history[k].add(v)

    def handle_history(self, record):
        self._dispatch_record(record)
        self._save_history(record)
        history_dict = proto_util.dict_from_proto_list(record.history.item)
        self._consolidated_summary.update(history_dict)
        self._save_summary(self._consolidated_summary)

    def handle_summary(self, record):
        summary = record.summary

        for item in summary.update:
            if len(item.nested_key) > 0:
                # we use either key or nested_key -- not both
                assert item.key == ""
                key = tuple(item.nested_key)
            else:
                # no counter-assertion here, because technically
                # summary[""] is valid
                key = (item.key,)

            target = self._consolidated_summary

            # recurse down the dictionary structure:
            for prop in key[:-1]:
                target = target[prop]

            # use the last element of the key to write the leaf:
            target[key[-1]] = json.loads(item.value_json)

        for item in summary.remove:
            if len(item.nested_key) > 0:
                # we use either key or nested_key -- not both
                assert item.key == ""
                key = tuple(item.nested_key)
            else:
                # no counter-assertion here, because technically
                # summary[""] is valid
                key = (item.key,)

            target = self._consolidated_summary

            # recurse down the dictionary structure:
            for prop in key[:-1]:
                target = target[prop]

            # use the last element of the key to erase the leaf:
            del target[key[-1]]

        self._save_summary(self._consolidated_summary)

    def handle_exit(self, record):
        self._dispatch_record(record, always_send=True)

    def handle_final(self, record):
        self._dispatch_record(record, always_send=True)

    def handle_header(self, record):
        self._dispatch_record(record)

    def handle_footer(self, record):
        self._dispatch_record(record)

    def handle_request_check_version(self, record):
        self._dispatch_record(record)

    def handle_request_run_start(self, record):
        run_start = record.request.run_start
        assert run_start
        assert run_start.run

        if not self._settings._disable_stats:
            pid = os.getpid()
            self._system_stats = stats.SystemStats(pid=pid, interface=self._interface,)
            self._system_stats.start()

        if not self._settings._disable_meta:
            run_meta = meta.Meta(settings=self._settings, interface=self._interface,)
            run_meta.probe()
            run_meta.write()

        self._tb_watcher = tb_watcher.TBWatcher(
            self._settings, interface=self._interface, run_proto=run_start.run,
        )

        result = alectio_helpers_internal_pb2.Result(uuid=record.uuid)
        self._result_q.put(result)

    def handle_request_resume(self, data):
        if self._system_stats is not None:
            logger.info("starting system metrics thread")
            self._system_stats.start()

    def handle_request_pause(self, data):
        if self._system_stats is not None:
            logger.info("stopping system metrics thread")
            self._system_stats.shutdown()

    def handle_request_poll_exit(self, record):
        self._dispatch_record(record, always_send=True)

    def handle_request_status(self, record):
        self._dispatch_record(record)

    def handle_request_get_summary(self, data):
        result = alectio_helpers_internal_pb2.Result(uuid=data.uuid)
        for key, value in six.iteritems(self._consolidated_summary):
            item = alectio_helpers_internal_pb2.SummaryItem()
            item.key = key
            item.value_json = json.dumps(value)
            result.response.get_summary_response.item.append(item)
        self._result_q.put(result)

    def handle_tbrecord(self, record):
        logger.info("handling tbrecord: %s", record)
        if self._tb_watcher:
            tbrecord = record.tbrecord
            self._tb_watcher.add(tbrecord.log_dir, tbrecord.save)
        self._dispatch_record(record)

    def handle_request_sampled_history(self, data):
        result = alectio_helpers_internal_pb2.Result(uuid=data.uuid)
        for key, sampled in six.iteritems(self._sampled_history):
            item = alectio_helpers_internal_pb2.SampledHistoryItem()
            item.key = key
            values = sampled.get()
            if all(isinstance(i, numbers.Integral) for i in values):
                item.values_int.extend(values)
            elif all(isinstance(i, numbers.Real) for i in values):
                item.values_float.extend(values)
            result.response.sampled_history_response.item.append(item)
        self._result_q.put(result)

    def handle_request_shutdown(self, record):
        # TODO(jhr): should we drain things and stop new requests from coming in?
        result = alectio_helpers_internal_pb2.Result(uuid=record.uuid)
        self._result_q.put(result)
        self._stopped.set()

    def finish(self):
        logger.info("shutting down handler")
        if self._tb_watcher:
            self._tb_watcher.finish()



"""
classes related to data upload.
uploading is used to keep
"""
from alectio_sdk.tools.mutations import UPLOAD_PARTNER_IMAGE_MUTATION, UPLOAD_PARTNER_NUMERICAL_MUTATION, UPLOAD_PARTNER_TEXT_MUTATION
from gql import gql
import asyncio


class BaseDataUpload():
    def __init__(self, client):
        self._client = client


class NumericalDataUpload(BaseDataUpload):
    """
    upload numerical data
    """
    def __init__(self, client):
        super().__init__(client)


    async def upload_data(self, records, job_id):
        # upload numerical data.
        # TODO: review upload for numerical data
        # should upload one json ?
        variables = {
            'file': open(records, 'r'),
            'records': records,
            'jobId': job_id
        }
        response = await self._client.execute(UPLOAD_PARTNER_NUMERICAL_MUTATION, variables=variables)
        print(await response.json())
        return

class ImageDataUpload(BaseDataUpload):
    """
    upload image data
    """
    def __init__(self, client):
        super().__init__(client)

    async def upload_data(self, records, job_id):
        # upload all the images asynchronously ...
        variables = {
            'files': [open(record['path'], 'rb') for record in records],
            'records': records,
            'jobId': job_id
        }
        response = await self._client.execute(UPLOAD_PARTNER_IMAGE_MUTATION, variables=variables)
        print(await response.json())
        return None

class TextDataUpload(BaseDataUpload):
    """
    upload text data
    """
    def __init__(self, client):
        super().__init__(client)


    async def upload_data(self, records, job_id):

        # upload text data.
        variables = {
            'file': open(records, 'r'),
            'records': records,
            'jobId': job_id
        }
        response = await self._client.execute(UPLOAD_PARTNER_TEXT_MUTATION, variables=variables)
        print(await response.json())
        return None



# create jobs once all the data has been sent.



import argparse
import copy
import datetime
from functools import wraps
import getpass
import json
import logging
import os
from os import path
import shutil
import subprocess
import sys
import textwrap
import time
import traceback
import xtarfile as tarfile
import click
from click.exceptions import ClickException
import six
from six.moves import configparser
import yaml
import alectio_sdk
import requests
from alectio_sdk.cli.alectio_cli import AlectioClient
from alectio_sdk.cli.datasets.downloader import Downloader
from tabulate import tabulate
import urllib.request
from os.path import expanduser
import os
import yaml


LOG_STRING = click.style("alectio", fg="blue", bold=True)
LOG_STRING_NOCOLOR = "alectio"
ERROR_STRING = click.style("ERROR", bg="red", fg="green")
WARN_STRING = click.style("WARNING", fg="yellow")
PRINTED_MESSAGES = set()

_silent = False
_show_info = True
_show_warnings = True
_show_errors = True
_logger = None

CONTEXT = dict(default_map={})

downloader = Downloader('universal')
def termerror(string, **kwargs):
    string = "\n".join(["{} {}".format(ERROR_STRING, s) for s in string.split("\n")])
    _log(
        string=string,
        newline=True,
        silent=not _show_errors,
        level=logging.ERROR,
        **kwargs
    )


def termwarn(string, **kwargs):
    string = "\n".join(["{} {}".format(WARN_STRING, s) for s in string.split("\n")])
    _log(
        string=string,
        newline=True,
        silent=not _show_errors,
        level=logging.WARN,
        **kwargs
    )

def _log(
    string="", newline=True, repeat=True, prefix=True, silent=False, level=logging.INFO
):
    global _logger
    silent = silent or _silent
    if string:
        if prefix:
            line = "\n".join(
                ["{}: {}".format(LOG_STRING, s) for s in string.split("\n")]
            )
        else:
            line = string
    else:
        line = ""
    if not repeat and line in PRINTED_MESSAGES:
        return
    # Repeated line tracking limited to 1k messages
    if len(PRINTED_MESSAGES) < 1000:
        PRINTED_MESSAGES.add(line)
    if silent:
        if level == logging.ERROR:
            logging.error(line)
        elif level == logging.WARNING:
            logging.warning(line)
        else:
            logging.info(line)
    else:
        click.echo(line, file=sys.stderr, nl=newline)

def cli_unsupported(argument):
    termerror("Unsupported argument `{}`".format(argument))
    sys.exit(1)


# class ClickWandbException(ClickException):
#     def format_message(self):
#         # log_file = util.get_log_file_path()
#         log_file = ""
#         orig_type = "{}.{}".format(self.orig_type.__module__, self.orig_type.__name__)
#         if issubclass(self.orig_type, Error):
#             return click.style(str(self.message), fg="red")
#         else:
#             return "An Exception was raised, see %s for full traceback.\n" "%s: %s" % (
#                 log_file,
#                 orig_type,
#                 self.message,
#             )

def display_error(func):
    """Function decorator for catching common errors and re-raising as wandb.Error"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            logging.error("".join(lines))

    return wrapper
class RunGroup(click.Group):
    @display_error
    def get_command(self, ctx, cmd_name):
        # TODO: check if cmd_name is a file in the current dir and not require `run`?
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        return None


def pretty(d, indent=0):
   for key, value in d.items():
      print('\t' * indent + str(key))
      if isinstance(value, dict):
         pretty(value, indent+1)
      else:
         print('\t' * (indent+1) + str(value))

## Click commands

@click.group()
@click.version_option(version=alectio_sdk.__version__)
@click.pass_context
def main(ctx):
    pass
    # if ctx.invoked_subcommand is None:
    #     click.echo(ctx.get_help())

@main.command(name='login')
@click.argument("key", nargs=-1)
@click.option(
    "--relogin", default=None, is_flag=True, help="Force relogin if already logged in."
)
@click.option("--anonymously", default=False, is_flag=True, help="Log in anonymously")
def login(key, relogin, anonymously, no_offline=False):

    if len(key) == 0:
        termerror("API Key is required.")
        sys.exit(0)

    key = key[0]
    
    headers = {"Authorization": "Bearer " + key}
    response = requests.post('https://api.alectio.com/api/me', headers=headers)
    if response.status_code == 401:
        termerror("API Key has expired or malformed")
    elif response.status_code == 200:
        payload = response.json()
        payload['api_key'] = key
        payload['user_id'] = payload['id']
        dict_file = [payload]

        alectio_dir = expanduser("~/.alectio")
        credential_file = expanduser(alectio_dir+'/credentials.yaml')
        if os.path.isdir(alectio_dir) and os.path.isfile(credential_file):
            termerror("Auth data already exists. If you want to reinitialize run alectio login with --relogin flag.")
        else:
            os.mkdir(alectio_dir) if not os.path.isdir(alectio_dir) else _log('Alectio dir exists')
            try:
                with open(credential_file, 'w') as file:
                    yaml.dump(dict_file, file)
            except:
                termerror('Not able to wirte file at ' +  credential_file)
            _log('Alectio Login complete')
    else:
        termerror("Something went wrong double check you API Key")

    _log('You are logged in as: ' + payload['username'])



@main.command(name='get')
@click.argument("entity", nargs=-1)
@click.option("--project", default=None, help="Specify project id, not sure of project id run 'alectio get projects'")
def get(entity, project):
    entity = entity[0]
    alectio = AlectioClient()
    if entity == 'projects':
        projects = alectio.projects()
        projects_table = []

        for project in projects:
            projects_table.append([project._name, project._id])        
        print(tabulate(projects_table, headers=['Name', 'Project ID'], tablefmt='fancy_grid'))
    
    if entity == 'project':
        if project is None:
            termerror("Project argument use 'alectio get project --project=PROJECT_ID'")
            sys.exit(1)
        else:
            project_data = alectio.project(project)
            print(project_data._name)
            print(project_data._attr['type'])
            print(project_data._id)
            
            experiments = project_data.experiments()
            for experiment in experiments:
                print(experiment._attr)
                print(experiment._name)
                print(experiment._id)


    if entity == 'experiments':
        experiments_table = []
        if project is None:
            termerror("Project argument use 'alectio get experiments --project=PROJECT_ID'")
            sys.exit(1)
        else:
            experiments = alectio.experiments(project)
            for experiment in experiments:
                experiments_table.append([experiment._name, experiment._id])        
                print(tabulate(experiments_table, headers=['Name', 'Experiment ID'], tablefmt='fancy_grid'))

    if entity == 'experiment':
        pass


@main.command(name='library-run')
@click.option("--token", default=None, help="Enter the Token you after creating an experiment for alectio library project")
@click.option("--resume", is_flag=True, default=False, help="Enter the Token you after creating an experiment for alectio library project")
@click.option("--auto_run", is_flag=True, default=False, help="Enter the Token you after creating an experiment for alectio library project")
def library_run(token, resume, auto_run):
    if path.exists('./.alectio') and not resume:
        termwarn('An experiment initialized here. Use either alectio library-run --resume or start a new experiment with alectio clear')
        sys.exit(0)
    elif path.exists('./.alectio') and resume:
        pass

    else:
        if token is None:
            termerror("Token is required.")
            sys.exit(1)
        _log("Experiment Initilizing.")

        if not resume:
            payload = requests.post('https://api.alectio.com/experiments/libraryRunFetch', json={'token': token}).json()
        else:
            payload_file = open('./.alectio/alectio_env.json')
            payload = json.load(payload_file)
            payload_file.close()

        if payload['status'] == 'error':
            termerror(payload['message'])
            sys.exit(1)
        
        if payload['data_url'] is None or payload['code_url'] is None:
            termerror('This Dataset or model is not supported it. Please check back')
            sys.exit(1)

        else:
            if not path.exists('./.alectio'):
                os.mkdir('./.alectio')
                with open('./.alectio/alectio_env.json', 'w') as fp:
                    json.dump(payload, fp)
            
            while True:
                if path.exists('./'+payload['code_file']):
                    break
                try:
                    urllib.request.urlretrieve(payload['code_url'], payload['code_file'])
                except Exception:
                    urllib.request.urlretrieve(payload['code_url'], payload['code_file'])
                else:
                    break


            # while True:

            #     response = requests.get(payload['code_url'], verify=False, headers={'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 11_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36'}, stream=True)
            #     if response.status_code == 200:
            #         with open(payload['code_file'], 'wb') as f:
            #             f.write(response.raw.read())
            #     else:
            #         if download == 2:
            #             termerror('Unable to get model. Please contact info@alectio.com if this persists')
            #             sys.exit(1)
            #         else:
            #             termerror('Retrying to fetch the model')
            #     download += 1
            if not path.exists('./'+payload['code_file'].replace('.tar.gz', "")):
                with tarfile.open(payload['code_file'], 'r') as archive:
                    archive.extractall()

            os.chdir(payload['code_file'].replace('.tar.gz', ''))
            if not path.exists('./data'):
                os.mkdir('data')
    
            if not path.exists('./log'):
                os.mkdir('log')

            if not path.exists('./weights'):
                os.mkdir('weights')
            
            if not path.exists('./weight'):
                os.mkdir('weight')

            if payload['data_url'] == 'Inplace':
                _log('Data will be dowloaded on first run of code')
        
            elif payload['data_url'] == 'Internal':
                pass

            else:
                while True:
                    if path.exists('./data/'+payload['data_file']):
                        break
                    try:
                        urllib.request.urlretrieve(payload['data_url'], "./data/"+payload['data_file'])
                    except Exception:
                        urllib.request.urlretrieve(payload['data_url'], "./data/"+payload['data_file'])
                    else:
                        break
                # downloader.download_file(payload['data_url'], './data' + payload['data_file'])
                _log('Extracting Data into data dir')
                os.chdir('./data')
                if not path.exists('./'+payload['data_file'].replace('.tar.gz', "")):
                    with tarfile.open('./' + payload['data_file'], 'r') as archive:
                        archive.extractall()
                os.chdir('../')

                _log('All files are feteched.')

                if auto_run:
                    pass
                else:
                    print('Run you experiment by installing all dependencies with %s/requirements.txt and the run running python main.py %s at: %s'  % (os.getcwd(), token, os.getcwd()))




        








class ALExperiment():
    def __init__(self, id):
        self.id = id


    def upload_record_db(self, path_to_file):
        "upload index to record mapping"
        return

    def upload_inference_stats(self, path_to_file):
        "upload final layer nn layer, need unlabled pool"
        return

    def upload_and_compute_metrics(self):
        """
        compute metrics and upload metrics to backend
        """
        return

    def unlabeled_pool(self):
        """
        retreive unlabeled pool to perform inference on
        """
        return





from pathlib import Path
import logging
import argparse
from botocore.config import Config

from boto3 import Session
import requests


logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class S3MultipartUploadUtil:
    """
    AWS S3 Multipart Upload Uril
    """
    def __init__(self, session: Session):
        self.session = session
        self.s3 = session.client('s3', config=Config(signature_version='s3v4'))
        self.upload_id = None
        self.bucket_name = None
        self.key = None

    def start(self, bucket_name: str, key: str):
        """
        Start Multipart Upload
        :param bucket_name:
        :param key:
        :return:
        """
        self.bucket_name = bucket_name
        self.key = key
        res = self.s3.create_multipart_upload(Bucket=bucket_name, Key=key)
        self.upload_id = res['UploadId']
        logger.debug(f"Start multipart upload '{self.upload_id}'")

    def create_presigned_url(self, part_no: int, expire: int=3600) -> str:
        """
        Create pre-signed URL for upload part.
        :param part_no:
        :param expire:
        :return:
        """
        signed_url = self.s3.generate_presigned_url(
            ClientMethod='upload_part',
            Params={'Bucket': self.bucket_name,
                    'Key': self.key,
                    'UploadId': self.upload_id,
                    'PartNumber': part_no},
            ExpiresIn=expire)
        logger.debug(f"Create presigned url for upload part '{signed_url}'")
        return signed_url

    def complete(self, parts):
        """
        Complete Multipart Uploading.
        `parts` is list of dictionary below.
        ```
        [ {'ETag': etag, 'PartNumber': 1}, {'ETag': etag, 'PartNumber': 2}, ... ]
        ```
        you can get `ETag` from upload part response header.
        :param parts: Sent part info.
        :return:
        """
        res = self.s3.complete_multipart_upload(
            Bucket=self.bucket_name,
            Key=self.key,
            MultipartUpload={
                'Parts': parts
            },
            UploadId=self.upload_id
        )
        logger.debug(f"Complete multipart upload '{self.upload_id}'")
        logger.debug(res)
        self.upload_id = None
        self.bucket_name = None
        self.key = None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('target_file')
    parser.add_argument('--bucket', required=True)
    args = parser.parse_args()

    target_file = Path(args.target_file)
    print(target_file)
    bucket_name = args.bucket
    key = target_file.name
    max_size = 5 * 1024 * 1024

    file_size = target_file.stat().st_size
    upload_by = int(file_size / max_size) + 1

    session = Session()
    s3util = S3MultipartUploadUtil(session)

    s3util.start(bucket_name, "0171cb0e02d811ebbda48c85900d6ee2/"+key)
    urls = []
    for part in range(1, upload_by + 1):
        signed_url = s3util.create_presigned_url(part)
        urls.append(signed_url)

    parts = []
    with target_file.open('rb') as fin:
        for num, url in enumerate(urls):
            part = num + 1
            file_data = fin.read(max_size)
            print(f"upload part {part} size={len(file_data)}")
            res = requests.put(url, data=file_data)
            print(res)
            if res.status_code != 200:
                return
            etag = res.headers['ETag']
            parts.append({'ETag': etag, 'PartNumber': part})

    print(parts)

    session1 = Session()
    s3util1 = S3MultipartUploadUtil(session1)
    s3util1.start(bucket_name, "0171cb0e02d811ebbda48c85900d6ee2/"+key)
    s3util1.upload_id = s3util.upload_id
    s3util1.complete(parts)


if __name__ == '__main__':
    main()


from os.path import expanduser
import os
import yaml
def extract_id(attr):
    """
    extract id from the attribute
    :params: attr primary key or sort key  
    """
    id = "".join(attr.split("#")[-1])
    return id


def credentials_exists():
    alectio_dir = expanduser("~/.alectio")
    credential_file = expanduser(alectio_dir+'/credentials.yaml')
    credential_data = None
    if os.path.isdir(alectio_dir) and os.path.isfile(credential_file):
        with open(credential_file, 'r') as stream:
            try:
                credential_data = yaml.safe_load(stream)[0]
            except yaml.YAMLError as exc:
                print(exc)
                print('\n Alectio Error: credentials.yaml is malformed\n')
                exit(1)
    os.environ['ALECTIO_API_KEY'] = credential_data['api_key']
    return True

def get_credentials():
    alectio_dir = expanduser("~/.alectio")
    credential_file = expanduser(alectio_dir+'/credentials.yaml')
    credential_data = None
    if os.path.isdir(alectio_dir) and os.path.isfile(credential_file):
        with open(credential_file, 'r') as stream:
            try:
                credential_data = yaml.safe_load(stream)[0]
                os.environ['ALECTIO_API_KEY'] = credential_data['api_key']
            except yaml.YAMLError as exc:
                print(exc)
                print('\n Alectio Error: credentials.yaml is malformed\n')
                exit(1)
    else:
        print('\n Alectio Error: credentials.yaml not found. Run alectio login YOUR_API_KEY\n')
        exit(1)
    return credential_data





from flask import jsonify
from flask import Flask, Response
from flask import request
from flask import send_file

# from waitress import serve
import numpy as np
import json
import requests
import os
import psutil
import time
import boto3
import logging
from copy import deepcopy
from .s3_client import S3Client
from alectio_sdk.metrics.object_detection import Metrics, batch_to_numpy
from alectio_sdk.metrics.classification import ClassificationMetrics
from alectio_sdk.metrics.object_segmentation import segmentation_metrics
from alectio_sdk.metrics.multilabel_classification import MultiLabelClassificationMetrics
from alectio_sdk.metrics.object_segmentation import SegMetrics
from alectio_sdk.sdk.sql_client import create_connection
from alectio_sdk.backend.backend import Backend

# modules for testing
import argparse
import json


class Pipeline(object):
    r"""
    A wrapper for your `train`, `test`, and `infer` function. The arguments for your functions should be specifed
    separately and passed to your pipeline object during creation.

    Args:
        name (str): experiment name
        train_fn (function): function to be executed in the train cycle of the experiment.
        test_fn (function): function to be executed in the test cycle of the experiment.
        infer_fn (function): function to be executed in the inference cycle of the experiment.
        getstate_fn (function): function specifying a mapping between indices and file names.

    """

    def __init__(self, name, train_fn, test_fn, infer_fn, getstate_fn, args, token):
        """
        sentry_sdk.init(
            dsn="https://4eedcc29fa7844828397dca4afc2db32@o409542.ingest.sentry.io/5282336",
            integrations=[FlaskIntegration()]
        )
        """
        self.app = Flask(name)
        self.gunicorn_logger = logging.getLogger("gunicorn.error")
        self.app.logger.handlers = self.gunicorn_logger.handlers
        self.app.logger.setLevel(self.gunicorn_logger.level)
        self.train_fn = train_fn
        self.test_fn = test_fn
        self.infer_fn = infer_fn
        self.getstate_fn = getstate_fn
        self.args = args
        self.client = S3Client()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(dir_path, "config.json"), "r") as f:
            self.config = json.load(f)

        self.client_token = token
        self.backend = None
        self.labeled = []

    def _notifyserverstatus(self, logdir):
        logging.basicConfig(
            filename=os.path.join(logdir, "Appstatus.log"), level=logging.INFO
        )
        self.app.logger.info("Flask app from Alectio initialized successfully")
        self.app.logger.info(
            "Training checkpoints and other logs for current experiment will be written into the folder {}".format(
                logdir
            )
        )
        self.app.logger.info(
            "Press CTRL + C to exit flask app , if Flask app terminates in the middle use fuser -k <port number>/tcp to terminate current process and relaunch alectio sdk"
        )

    def _checkdirs(self, dir_):
        if not os.path.exists(dir_):
            os.makedirs(dir_, exist_ok=True)

    def check_model_activations(self, model_outputs):
        supported_activations = {'pre_softmax', 'softmax', 'sigmoid', 'logits'}
        activation_set = set()
        for k, v in model_outputs.items():
            if isinstance(v, dict):
                activation_set.update(set(v.keys()).intersection(supported_activations))
            if len(activation_set) == 0:
                raise TypeError('Incorrect logit output format encountered')

        return activation_set

    def _estimate_exp_time(self, last_time):
        """
        Estimates the compute time remaining for the experiment

        Args:
            train_times (list): training_times noted down so far
            n_loop (int): total number of loops
        """

        def convert(seconds):
            seconds = seconds % (24 * 3600)
            hour = seconds // 3600
            seconds %= 3600
            minutes = seconds // 60
            seconds %= 60

            return "%d:%02d:%02d" % (hour, minutes, seconds)

        loops_completed = self.cur_loop + 1
        time_left = convert(last_time * (self.n_loop - loops_completed))
        self.app.logger.info("Estimated time left for the experiment: {}".format(time_left))
        return time_left

    def one_loop(self, request):
        # Get payload args

        self.app.logger.info("Extracting payload arguments from Alectio")
        # Get payload args
        payload = {
            "experiment_id": request["experiment_id"],
            "project_id": request["project_id"],
            "cur_loop": request["cur_loop"],
            "user_id": request["user_id"],
            "bucket_name": request["bucket_name"],
            "type": request["type"],
            "n_rec": request["n_rec"],
            "n_loop": request["n_loop"]
        }

        self.logdir = payload["experiment_id"]
        self._checkdirs(self.logdir)

        self._notifyserverstatus(self.logdir)
        self.app.logger.info("Valid payload arguments extracted")
        self.app.logger.info("Initializing process to train and optimize your model")
        returned_payload = self._one_loop(payload, self.args)
        self.app.logger.info("Optimization process complete !")
        self.app.logger.info(
            "Your results for this loop should be visible in Alectio website shortly"
        )

        backend_ip = self.config["backend_ip"]
        port = 80
        url = "".join(["http://", backend_ip, ":{}".format(port), "/end_of_task"])

        headers = {"Authorization": "Bearer " + self.client_token}
        status = requests.post(
            url=url, json=returned_payload, headers=headers
        ).status_code
        if status == 200:
            self.app.logger.info(
                "Experiment {} running".format(payload["experiment_id"])
            )
            return {"Message": "Loop Started - 200 status returned"}
        else:
            return {"Message": "Loop Failed - non 200 status returned"}

    def _one_loop(self, payload, args):
        r"""
        Executes one loop of active learning. Returns the read `payload` back to the user.

        Args:
           args: a dict with the key `sample_payload` (required path) and any arguments needed by the `train`, `test`
           and infer functions.
        Example::

            args = {sample_payload: 'sample_payload.json', EXPT_DIR : "./log", exp_name: "test", \
                                                                 train_epochs: 1, batch_size: 8}
            app._one_loop(args)

        """

        # payload = json.load(open(args["sample_payload"]))
        self.app.logger.info("Extracting essential experiment params")

        # read selected indices upto this loop
        payload["cur_loop"] = int(payload["cur_loop"])
        # self.curout_loop = payload["cur_loop"]

        # Leave cur_loop - 1 when doing a 1 dag solution, when doing 2 dag cur_loop remains the same
        self.cur_loop = payload["cur_loop"]
        self.bucket_name = payload["bucket_name"]
        self.n_loop = payload["n_loop"]

        # type of the ML problem

        self.type = payload["type"]
        self.type = self.type.lower()
        print('Task type: {}'.format(self.type))
        # dir for expt log in S3
        expt_dir = [payload["user_id"], payload["project_id"], payload["experiment_id"]]

        if self.bucket_name == self.config["sandbox_bucket"]:
            # shared S3 bucket for sandbox user
            self.expt_dir = os.path.join(
                payload["user_id"], payload["project_id"], payload["experiment_id"]
            )

            self.project_dir = os.path.join(payload["user_id"], payload["project_id"])

        else:
            # dedicated S3 bucket for paid user
            self.expt_dir = os.path.join(
                payload["project_id"], payload["experiment_id"]
            )

            self.project_dir = os.path.join(payload["project_id"])

        self.app.logger.info("Essential experiment params have been extracted")
        # get meta-data of the data set
        self.app.logger.info("Verifying the meta.json that was uploaded by the user")
        key = os.path.join(self.project_dir, "meta.json")
        bucket = boto3.resource("s3").Bucket(self.bucket_name)

        json_load_s3 = lambda f: json.load(bucket.Object(key=f).get()["Body"])
        self.meta_data = json_load_s3(key)
        self.app.logger.info(
            "SDK Retrieved file: {} from bucket : {}".format(key, self.bucket_name)
        )

        # self.meta_data = self.client.read(self.bucket_name, key, "json")
        # logging.info('SDK Retrieved file: {} from bucket : {}'.format(key, self.bucket_name))

        self.labeled = []
        self.app.logger.info("Reading indices to train on")

        for i in range(self.cur_loop + 1):
            object_key = os.path.join(self.expt_dir, "selected_indices_{}.pkl".format(i))
            selected_indices = self.client.read(self.bucket_name, object_key=object_key, file_format="pickle")
            self.labeled.extend(selected_indices)

        if self.cur_loop == 0:

            self.resume_from = None
            self.app.logger.info(
                "Extracting indices for our reference, this may take time ... Please be patient"
            )
            self.state_json = self.getstate_fn(args)
            object_key = os.path.join(self.expt_dir, "data_map.pkl")
            self.app.logger.info("Extraction complete !!!")
            self.app.logger.info(
                "Creating index to records data reference for the current experiment"
            )
            self.client.multi_part_upload_with_s3(
                self.state_json, self.bucket_name, object_key, "pickle"
            )
            self.app.logger.info("Reference creation complete")

        else:

            # check if ckpt cur_loop - 1 exists, otherwise we need to download it from S3
            # if not os.path.isfile(
            #         os.path.join(self.args["EXPT_DIR"], f"ckpt_{self.cur_loop - 1}")
            # ):
            #     # need to download the checkpoint files from S3
            #     self.app.logger.info(
            #         "Starting to copy checkpoints for cloned experiment..."
            #     )
            #     self.client.download_checkpoints(
            #         payload["bucket_name"],
            #         payload["project_id"],
            #         payload["experiment_id"],
            #         payload["cur_loop"],
            #         self.args["EXPT_DIR"],
            #     )
            #     self.app.logger.info(
            #         "Finished downloading checkpoints for cloned experiment"
            #     )

            self.app.logger.info("Resuming from a checkpoint from a previous loop ")
            # two dag approach needs to refer to the previous checkpoint
            self.resume_from = "ckpt_{}".format(self.cur_loop - 1)

        self.ckpt_file = "ckpt_{}".format(self.cur_loop)
        self.app.logger.info("Initializing training of your model")

        self.train(args)
        self.app.logger.info("Training complete !")
        self.app.logger.info("Initializing testing of your model !")
        self.test(args)
        self.app.logger.info("Testing complete !")
        self.app.logger.info("Assessing current best model")
        self.infer(args)
        self.app.logger.info("Assessment complete ")
        self.app.logger.info(
            "Time to check what records to train on next loop , visit our front end for more details"
        )

        # Drop unwanted payload values
        del payload["type"]
        del payload["cur_loop"]
        del payload["bucket_name"]
        return payload

    def train(self, args):
        r"""
        A wrapper for your `train` function. Returns `None`.

        Args:
           args: a dict whose keys include all of the arguments needed for your `train` function which is defined in `processes.py`.

        """
        start = time.time()

        if self.labeled is not None and len(self.labeled) == 0:
            raise ValueError('Labeled indices from backend are empty or None')

        self.app.logger.info("Labeled records are ready to be trained")
        self.labeled.sort()  # Maintain increasing order

        # if len(self.labeled) > len(set(self.labeled)):
        #     raise ValueError("There exist repeated records.")
        from collections import Counter
        all_labeled = dict(Counter(self.labeled))
        duplicates = []
        for k, v in all_labeled.items():
            if v == 2:
                duplicates.append(k)

        self.labeled = list(set(self.labeled))

        train_op = self.train_fn(
            args,
            labeled=deepcopy(self.labeled),
            resume_from=self.resume_from,
            ckpt_file=self.ckpt_file,
        )

        if train_op is not None:
            labels = train_op["labels"]
            unique, counts = np.unique(labels, return_counts=True)
            num_queried_per_class = {u: c for u, c in zip(unique, counts)}

        end = time.time()

        # @TODO compute insights from labels
        if train_op is not None:
            insights = {"train_time": end - start, "num_queried_per_class": num_queried_per_class}
        else:
            insights = {"train_time": end - start}

        self._estimate_exp_time(insights["train_time"])
        object_key = os.path.join(
            self.expt_dir, "insights_{}.pkl".format(self.cur_loop)
        )

        self.client.multi_part_upload_with_s3(
            insights, self.bucket_name, object_key, "pickle"
        )

        return

    def test(self, args):
        r"""
        A wrapper for your `test` function which writes predictions and ground truth to the specified S3 bucket. Returns `None`.

        Args:
           args: a dict whose keys include all of the arguments needed for your `test` function which is defined in `processes.py`.

        """
        self.app.logger.info("Extracting test results ")
        res = self.test_fn(args, ckpt_file=self.ckpt_file)

        if self.type == "3d_segmentation":
            if "rangebased" in res:
                _3D_predictions, _3D_labels, _2D_predictions, _2D_labels, range_filtered, range_default = \
                    res["3Dpredictions"], res["3Dlabels"], res["2Dpredictions"], res["2Dlabels"], \
                    res["rangebased"], res["default"]
            else:
                _3D_predictions, _3D_labels, _2D_predictions, _2D_labels, range_filtered, range_default = \
                    res["3Dpredictions"], res["3Dlabels"], res["2Dpredictions"], res["2Dlabels"], None, None
        else:
            predictions, ground_truth = res["predictions"], res["labels"]
        self.app.logger.info("Writing test results to S3")

        # write predictions and labels to S3
        object_key = os.path.join(
            self.expt_dir, "test_predictions_{}.pkl".format(self.cur_loop)
        )
        self.client.multi_part_upload_with_s3(
            predictions, self.bucket_name, object_key, "pickle"
        )

        if self.cur_loop == 0:
            # write ground truth to S3
            object_key = os.path.join(
                self.expt_dir, "test_ground_truth.pkl".format(self.cur_loop)
            )
            self.client.multi_part_upload_with_s3(
                ground_truth, self.bucket_name, object_key, "pickle"
            )

        if "3D" in self.type:
            self.compute3D_metrics(_3D_predictions, _3D_labels, _2D_predictions, _2D_labels, range_filtered)
        else:
            self.compute_metrics(predictions, ground_truth)
        return

    def compute3D_metrics(self, pclpredictions, pcllabels, imgpredictions, imglabels, range_filtered, range_default):
        metrics = {}
        if self.type == "3d_object_detection":
            raise NotImplementedError("3D object detection evaluation has not been implemented yet")

        elif self.type == "3d_segmentation":
            m = SegMetrics(
                n_classes=len(self.meta_data["class_labels"]),
                labels=self.meta_data["class_labels"],
                return_2D=True,
                return_3D=True,
                rangenet=True,
                default_ranges=range_default,
                include_ranges=True,
                range_labels=range_filtered,
            )

            m.evaluate3D(pcllabels, pclpredictions, rangelabels=range_filtered)
            m.evaluate2D(imglabels, imgpredictions)
            metrics = {
                "3DCM": m.get3DCM(),
                "3DrangeCM": m.get3DrangeCM(),
                "3DmIOU": m.get3DmIOU(),
                "3DrangemIOU": m.get3DrangemIOU(),
                "3DIOU": m.get3DIOU(),
                "3Drange3DIOU": m.get3DrangeIOU(),
                "3DDICE": m.get3DDICE(),
                "3DmDICE": m.get3DmDICE(),
                "3DrangeDICE": m.get3DrangeDICE(),
                "3DrangemDICE": m.get3DrangemDICE(),
                "3DAcc": m.get3Dacc(),
                "3DmAcc": m.get3Dmacc(),
                "3DrangeAcc": m.get3Drangeacc(),
                "3DrangemAcc": m.get3Drangemacc(),
                "class_labels": self.meta_data["class_labels"],
            }

    def compute_metrics(self, predictions, ground_truth):
        metrics = {}
        if self.type == "2d_object_detection":
            det_boxes, det_labels, det_scores, true_boxes, true_labels = batch_to_numpy(
                predictions, ground_truth
            )

            m = Metrics(
                det_boxes=det_boxes,
                det_labels=det_labels,
                det_scores=det_scores,
                true_boxes=true_boxes,
                true_labels=true_labels,
                num_classes=len(self.meta_data["class_labels"]),
            )

            metrics = {
                "mAP": m.getmAP(),
                "AP": m.getAP(),
                "precision": m.getprecision(),
                "recall": m.getrecall(),
                "confusion_matrix": m.getCM().tolist(),
                "class_labels": self.meta_data["class_labels"],
            }

        elif self.type == "text_classification" or self.type == "image_classification":

            m = ClassificationMetrics(predictions, ground_truth)

            metrics = {
                "accuracy": m.get_accuracy(),
                "f1_score_per_class": m.get_f1_score_per_class(),
                "f1_score": m.get_f1_score(),
                "precision_per_class": m.get_precision_per_class(),
                "precision": m.get_precision(),
                "recall_per_class": m.get_recall_per_class(),
                "recall": m.get_recall(),
                "confusion_matrix": m.get_confusion_matrix(),
                "acc_per_class": m.get_acc_per_class(),
                "label_disagreement": m.get_label_disagreement(),
            }

        elif self.type == "2d_segmentation":

            metrics = segmentation_metrics(ground_truth, predictions, n_classes=len(self.meta_data["class_labels"]))

        elif self.type == "multilabel_text_classification" or self.type == "multi_label_text_classification":

            m = MultiLabelClassificationMetrics(predictions, ground_truth)

            metrics = {
                "accuracy": m.get_accuracy(),
                "micro_f1": m.get_f1_score_micro(),
                "macro_f1": m.get_f1_score_macro(),
                "micro_precision": m.get_precision_micro(),
                "macro_precision": m.get_precision_macro(),
                "micro_recall": m.get_recall_micro(),
                "macro_recall": m.get_recall_macro(),
                "multilabel_confusion_matrix": m.get_confusion_matrix(),
                "hamming_loss": m.get_hamming_loss()
            }

        else:
            raise ValueError("The selected task type is currently not supported, received type : {}".format(self.type))

        # save metrics to S3
        object_key = os.path.join(self.expt_dir, "metrics_{}.pkl".format(self.cur_loop))
        self.client.multi_part_upload_with_s3(
            metrics, self.bucket_name, object_key, "pickle"
        )

        return

    def infer(self, args):
        r"""
        A wrapper for your `infer` function which writes outputs to the specified S3 bucket. Returns `None`.

        Args:
           args: a dict whose keys include all of the arguments needed for your `infer` function which is defined in `processes.py`.

        """
        self.app.logger.info(
            "Getting insights on currently unused/unlabelled train data"
        )
        self.app.logger.warning(
            "This may take some time. Please be patient ............"
        )

        ts = range(self.meta_data["train_size"])
        self.unlabeled = sorted(list(set(ts) - set(self.labeled)))
        args['cur_loop'] = self.cur_loop
        outputs = self.infer_fn(
            args, unlabeled=deepcopy(self.unlabeled), ckpt_file=self.ckpt_file
        )
        local_db = os.path.join(self.args['EXPT_DIR'], 'infer_outputs_{}.db'.format(self.cur_loop))

        if os.path.exists(local_db) or outputs is not None:
            self.app.logger.info(
                "Sending assessments on unlabelled train set to Alectio team"
            )
            if "classification" in self.type:
                outputs = outputs['outputs']
                # finds correct activation as key i.e. 'sigmoid' or 'logits'
                activation = self.check_model_activations(outputs)

                # Remap to absolute indices
                logits_dict = {}
                remap_dict = {}
                for key in outputs.keys():
                    remap_dict[key] = self.unlabeled.pop(0)

                while len(outputs) > 0:
                    for orig_ix, correct_ix in remap_dict.items():
                        val = outputs.pop(orig_ix)
                        logits_dict[correct_ix] = {}
                        for a in activation:
                            logits_dict[correct_ix][a] = val[a]
                object_key = os.path.join(self.expt_dir, 'logits_{}.pkl'.format(self.cur_loop))
                self.client.multi_part_upload_with_s3(
                    logits_dict,
                    self.bucket_name,
                    object_key,
                    "pickle"
                )
            elif "object_detection" or "semantic" in self.type:
                object_key = os.path.join(self.expt_dir, 'infer_outputs_{}.db'.format(self.cur_loop))
                logits_conn = create_connection(local_db)

                if logits_conn is not None:
                    logits_cur = logits_conn.cursor()
                    sql = """
                            UPDATE indexes
                            SET row_id = ?
                            WHERE id = ?
                    """
                    for index, correct_ix in enumerate(self.unlabeled):
                        logits_cur.execute(sql, (correct_ix, index))

                    logits_conn.close()
                else:
                    raise ConnectionError('Cannot connect to softmax database')

                # upload sqlite db
                self.client.multi_part_upload_file(local_db, self.bucket_name, object_key)
                # delete local version
                os.remove(local_db)
            else:
                raise ValueError('Use case not found')
        else:
            raise ValueError('Infer outputs not returned. You must return outputs from the implemented infer() '
                             'function. Please refer to the README on the Alectio SDK repository.')

        return

    def __call__(self, debug=False, host="0.0.0.0", port=80):
        r"""
        A wrapper for your `test` function which writes predictions and ground truth to the specified S3 bucket. Returns `None`.

        Args:
           debug (boolean, Default=False): If set to true, then the app runs in debug mode. See https://flask.palletsprojects.com/en/1.1.x/api/#flask.Flask.debug.
           host (str, Default='0.0.0.0'): the hostname to be listened to.
           port(int, Default:5000): the port of the webserver.

        """
        # As of now SDK will start the experiment everytime the SDK is initlized
        # TODO:: ADD RESUME OPTION

        # print("Starting the Experiment. Alectio is fetching indices, please wait.")
        self.backend = Backend(self.client_token)

        backend_ip = self.config["backend_ip"]
        url = "".join(["http://", backend_ip, ":{}".format(port)])

        startExpURL = "".join([url, '/api/v0/start_experiment_response'])
        sdkResponseURL = "".join([url, '/api/v0/get_sdk_response'])
        headers = {"Authorization": "Bearer " + self.client_token}

        response = requests.post(url=startExpURL, json={'exp_token': self.client_token})
        print(response.json()['message'])
        count = 0
        if response.json()['message'] == "Started":
            while True:
                response_child = requests.post(url=sdkResponseURL, json={'exp_token': self.client_token})
                if response_child.json()['status'] == "Fetched":
                    print('\n')
                    one_loop_response = self.one_loop(response_child.json())
                    print('\n')
                    print(one_loop_response)
                    count = 0
                if response_child.json()['status'] == "Finished":
                    print('\n')
                    print("Experiment complete")
                    os.environ["AWS_ACCESS_KEY_ID"] = " "
                    os.environ["AWS_SECRET_ACCESS_KEY"] = " "
                    break
                if response_child.json()['status'] == "Failed":
                    if count == 0:
                        print('\n')
                        print('Waiting for server.', end='', flush=True)
                        count += 1
                        time.sleep(10)
                    else:
                        time.sleep(10)
                        print(".", end='', flush=True)

    @staticmethod
    def shutdown_server():
        func = request.environ.get("werkzeug.server.shutdown")
        if func is None:
            raise RuntimeError("Not running with the Werkzeug Server")
        func()

    @staticmethod
    def end_exp():
        print()
        print("======== Experiment Ended ========")
        print("Server shutting down...")
        p = psutil.Process(os.getpid())
        p.terminate()
        return "Experiment complete"



import base64
import binascii
import collections
import logging
import threading
import requests
import time
import alectio_helpers
import itertools
from six.moves import queue
from alectio_helpers import util
from alectio_helpers import env
import os


MAX_LINE_SIZE = 4 * 1024 * 1024 - 100 * 1024  # imposed by back end

logger = logging.getLogger(__name__)

Chunk = collections.namedtuple("Chunk", ("filename", "data"))


class DefaultFilePolicy(object):
    def __init__(self, start_chunk_id=0):
        self._chunk_id = start_chunk_id

    def process_chunks(self, chunks):
        chunk_id = self._chunk_id
        self._chunk_id += len(chunks)
        return {"offset": chunk_id, "content": [c.data for c in chunks]}


class JsonlFilePolicy(DefaultFilePolicy):
    def process_chunks(self, chunks):
        chunk_id = self._chunk_id
        self._chunk_id += len(chunks)
        chunk_data = []
        for chunk in chunks:
            if len(chunk.data) > MAX_LINE_SIZE:
                msg = "Metric data exceeds maximum size of {} bytes. Dropping it.".format(
                    MAX_LINE_SIZE
                )
                alectio_helpers.termerror(msg, repeat=False)
                util.sentry_message(msg)
            else:
                chunk_data.append(chunk.data)

        return {
            "offset": chunk_id,
            "content": chunk_data,
        }


class SummaryFilePolicy(DefaultFilePolicy):
    def process_chunks(self, chunks):
        data = chunks[-1].data
        if len(data) > MAX_LINE_SIZE:
            msg = "Summary data exceeds maximum size of {} bytes. Dropping it.".format(
                MAX_LINE_SIZE
            )
            alectio_helpers.termerror(msg, repeat=False)
            util.sentry_message(msg)
            return False
        return {"offset": 0, "content": [data]}


class CRDedupeFilePolicy(DefaultFilePolicy):
    """File stream policy that removes characters that would be erased by
    carriage returns.

    This is what a terminal does. We use it for console output to reduce the
    amount of data we need to send over the network (eg. for progress bars),
    while preserving the output's appearance in the web app.
    """

    def process_chunks(self, chunks):
        ret = []
        flag = False  # whether the cursor can be moved up
        for c in chunks:
            # Line has two possible formats:
            # 1) "2020-08-25T20:38:36.895321 this is my line of text"
            # 2) "ERROR 2020-08-25T20:38:36.895321 this is my line of text"
            prefix = ""
            token, rest = c.data.split(" ", 1)
            if token == "ERROR":
                prefix += token + " "
                token, rest = rest.split(" ", 1)
            prefix += token + " "

            lines = rest.split(os.linesep)
            for line in lines:
                line = line.split("\r")[-1]
                if line:
                    # check for cursor up control character
                    if line.endswith("\x1b\x5b\x41"):
                        if flag:
                            ret.pop()
                            flag = False
                    else:
                        ret.append(prefix + line + os.linesep)
                        flag = True
        chunk_id = self._chunk_id
        self._chunk_id += len(ret)
        return {"offset": chunk_id, "content": ret}


class BinaryFilePolicy(DefaultFilePolicy):
    def process_chunks(self, chunks):
        data = b"".join([c.data for c in chunks])
        enc = base64.b64encode(data).decode("ascii")
        offset = self._offset
        self._offset += len(data)
        return {"offset": self._offset, "content": enc, "encoding": "base64"}


class FileStreamApi(object):
    """Pushes chunks of files to our streaming endpoint.

    This class is used as a singleton. It has a thread that serializes access to
    the streaming endpoint and performs rate-limiting and batching.

    TODO: Differentiate between binary/text encoding.
    """

    Finish = collections.namedtuple("Finish", ("exitcode"))

    HTTP_TIMEOUT = env.get_http_timeout(10)
    MAX_ITEMS_PER_PUSH = 10000

    def __init__(self, api, run_id, start_time, settings=None):
        if settings is None:
            settings = dict()
        self._settings = settings
        self._api = api
        self._run_id = run_id
        self._start_time = start_time
        self._client = requests.Session()
        self._client.auth = ("api", api.api_key)
        self._client.timeout = self.HTTP_TIMEOUT
        self._client.headers.update(
            {
                "User-Agent": api.user_agent,
                "X-alectio_helpers-USERNAME": env.get_username(),
                "X-alectio_helpers-USER-EMAIL": env.get_user_email(),
            }
        )
        self._file_policies = {}
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=self._thread_body)
        # It seems we need to make this a daemon thread to get sync.py's atexit handler to run, which
        # cleans this thread up.
        self._thread.daemon = True
        self._init_endpoint()

    def _init_endpoint(self):
        settings = self._api.settings()
        settings.update(self._settings)
        self._endpoint = "{base}/files/{entity}/{project}/{run}/file_stream".format(
            base=settings["base_url"],
            entity=settings["entity"],
            project=settings["project"],
            run=self._run_id,
        )

    def start(self):
        self._init_endpoint()
        self._thread.start()

    def set_default_file_policy(self, filename, file_policy):
        """Set an upload policy for a file unless one has already been set.
        """
        if filename not in self._file_policies:
            self._file_policies[filename] = file_policy

    def set_file_policy(self, filename, file_policy):
        self._file_policies[filename] = file_policy

    @property
    def heartbeat_seconds(self):
        # Defaults to 30
        return self._api.dynamic_settings["heartbeat_seconds"]

    def rate_limit_seconds(self):
        run_time = time.time() - self._start_time
        if run_time < 60:
            return max(1, self.heartbeat_seconds / 15)
        elif run_time < 300:
            return max(2.5, self.heartbeat_seconds / 3)
        else:
            return max(5, self.heartbeat_seconds)

    def _read_queue(self):
        # called from the push thread (_thread_body), this does an initial read
        # that'll block for up to rate_limit_seconds. Then it tries to read
        # as much out of the queue as it can. We do this because the http post
        # to the server happens within _thread_body, and can take longer than
        # our rate limit. So next time we get a chance to read the queue we want
        # read all the stuff that queue'd up since last time.
        #
        # If we have more than MAX_ITEMS_PER_PUSH in the queue then the push thread
        # will get behind and data will buffer up in the queue.
        return util.read_many_from_queue(
            self._queue, self.MAX_ITEMS_PER_PUSH, self.rate_limit_seconds()
        )

    def _thread_body(self):
        posted_data_time = time.time()
        posted_anything_time = time.time()
        ready_chunks = []
        finished = None
        while finished is None:
            items = self._read_queue()
            for item in items:
                if isinstance(item, self.Finish):
                    finished = item
                else:
                    # item is Chunk
                    ready_chunks.append(item)

            cur_time = time.time()

            if ready_chunks and (
                finished or cur_time - posted_data_time > self.rate_limit_seconds()
            ):
                posted_data_time = cur_time
                posted_anything_time = cur_time
                self._send(ready_chunks)
                ready_chunks = []

            if cur_time - posted_anything_time > self.heartbeat_seconds:
                posted_anything_time = cur_time
                self._handle_response(
                    util.request_with_retry(
                        self._client.post,
                        self._endpoint,
                        json={"complete": False, "failed": False},
                    )
                )
        # post the final close message. (item is self.Finish instance now)
        util.request_with_retry(
            self._client.post,
            self._endpoint,
            json={"complete": True, "exitcode": int(finished.exitcode)},
        )

    def _handle_response(self, response):
        """Logs dropped chunks and updates dynamic settings"""
        if isinstance(response, Exception):
            raise response
            alectio_helpers.termerror("Droppped streaming file chunk (see alectio_helpers/debug.log)")
            logging.error("dropped chunk %s" % response)
        elif response.json().get("limits"):
            parsed = response.json()
            self._api.dynamic_settings.update(parsed["limits"])

    def _send(self, chunks):
        # create files dict. dict of <filename: chunks> pairs where chunks is a list of
        # [chunk_id, chunk_data] tuples (as lists since this will be json).
        files = {}
        # Groupby needs group keys to be consecutive, so sort first.
        chunks.sort(key=lambda c: c.filename)
        for filename, file_chunks in itertools.groupby(chunks, lambda c: c.filename):
            file_chunks = list(file_chunks)  # groupby returns iterator
            # Specific file policies are set by internal/sender.py
            self.set_default_file_policy(filename, DefaultFilePolicy())
            files[filename] = self._file_policies[filename].process_chunks(file_chunks)
            if not files[filename]:
                del files[filename]

        self._handle_response(
            util.request_with_retry(
                self._client.post, self._endpoint, json={"files": files}
            )
        )

    def stream_file(self, path):
        name = path.split("/")[-1]
        with open(path) as f:
            self._send([Chunk(name, line) for line in f])

    def push(self, filename, data):
        """Push a chunk of a file to the streaming endpoint.

        Args:
            filename: Name of file that this is a chunk of.
            chunk_id: TODO: change to 'offset'
            chunk: File data.
        """
        self._queue.put(Chunk(filename, data))

    def finish(self, exitcode):
        """Cleans up.

        Anything pushed after finish will be dropped.

        Args:
            exitcode: The exitcode of the watched process.
        """
        self._queue.put(self.Finish(exitcode))
        self._thread.join()



import requests
import os
import json
import asyncio
import sys
import uuid
import yaml
import pathlib
from envyaml import EnvYAML
import os.path
from datetime import datetime

from gql import Client, gql
from aiogqlc import GraphQLClient

from gql.client import RetryError
from gql.transport.requests import RequestsHTTPTransport

from alectio_sdk.api.project import Project
from alectio_sdk.api.experiment import Experiment
from alectio_sdk.api.model import Model
from alectio_sdk.api.job import Job

from alectio_sdk.tools.utils import extract_id
from alectio_sdk.tools.fragments import *
from alectio_sdk.tools.mutations import *

from alectio_sdk.exceptions import APIKeyNotFound



#TODO: all plural objects should be iterables -  A.Y
class AlectioClient:
    def __init__(self, environ=os.environ):
        self._environ = environ


        if 'ALECTIO_API_KEY' not in self._environ:
            raise APIKeyNotFound

        self._api_key = self._environ['ALECTIO_API_KEY']
        # self._client_secret = self._environ['CLIENT_SECRET']
        # self._client_id = self._environ['CLIENT_ID']
        self._client_token = None

        # cli user settings
        self._settings = {
            'git_remote': "origin",
            'base_url': "https://api.alectio.com"
            #'base_url': "http://localhost:5005"
        }

        # self._endpoint = f'{self._settings['base_url']}/graphql'
        self._endpoint = self._settings['base_url'] + "/graphql"

        # graphql client
        self._client = Client(
            transport=RequestsHTTPTransport(
                url=self._endpoint,
                verify=False,
                retries=3,
            ),
            fetch_schema_from_transport=True,
        )

        # client to upload files, images, etc.
        # uses https://pypi.org/project/aiogqlc/
        self._upload_client = GraphQLClient('http://localhost:5005/graphql')
        self._oauth_server = 'http://a12a876383f3c493b9fb474f3ce1e5a4-2118098133.us-west-2.elb.amazonaws.com/'
        # need to retrive user_id based on token @ DEVI from OPENID
        self._user_id = "82b4fb909f1f11ea9d300242ac110002" # ideally this should be set already. Dummy one will be set when init is invoked
        # compnay id = 7774e1ca972811eaad5238c986352c36s
        # self.dir_path = os.path.dirname(os.path.realpath(__file__))

    def request_client_token(self):
        os.environ['client_id'] = input('Please Enter Clinet ID: ') # Will make this a env variable, Ideally this should be alredy set
        os.environ['client_secret'] = input('Please enter Clinet Secret: ')

        print('Opening OAuth URL to fetch credentials')

        ACCESS_URI = self._oauth_server + 'oauth/authorize?client_id=' + os.environ['client_id'] + '&scope=openid+profile&response_type=code&nonce=abc'
        run(ACCESS_URI)
        headers = {"Authorization": "Bearer " + self._client_token['access_token']}
        requests_data = requests.get(url=self._oauth_server + 'api/me', headers=headers)

        if requests_data.status_code == 200:
            requests_data = requests_data.json()
            self._user_id = requests_data['id']

    def init(self, file_path ='.alectio/token.json'):
        current_path = str(pathlib.Path(__file__).parent.absolute())
        file_path = current_path + '/' + file_path
        if file_path and self._client_token is None:
            if os.path.isfile(file_path):
                with open(file_path, 'r') as f:
                    self._client_token = json.load(f)
                headers = {"Authorization": "Bearer " + self._client_token['access_token']}
                requests_data = requests.get(
                    url=self._oauth_server + 'api/me', headers=headers)
                if requests_data.status_code == 401:
                    print("Clinet Token Expired. Fetching new one. Please follow steps")
                    self.request_client_token()
                elif requests_data.status_code == 200:
                    requests_data = requests_data.json()
                    self._user_id = requests_data['id']
                    print(self._user_id)
            else:
                requests_data = requests_data.json()

        #TODO: Create the token request if not present and save it to the disk


    def get_single(self, resource, query_string, params):
        """
        return a single object for the requested resource.
        :params: resource - name of the resource to obtain i.e experiments, projects, models, etc.
        :params: query_string - graphql string to invoke.
        :params: params - variables required to invoke query string in the client.
        """
        query = gql(query_string)
        class_name = lambda class_name: getattr(sys.modules[__name__], class_name)
        singular = self._client.execute(query, params)[resource][0]
        class_to_init = class_name(resource.title())
        hash_key =  extract_id(singular['pk'])
        if resource == "project":
           hash_key = extract_id(singular['sk'])

        if not resource == "job":
            singular_object = class_to_init(self._client, singular, self._user_id, hash_key)
            return singular_object

        # job object class is slighty different in design
        singular_object = class_to_init(self._upload_client, singular, hash_key)
        return singular_object


    def mutate_single(self, query_string, params):
        """
        return a single object for the requested resource.
        :params: resource - name of the resource to obtain i.e experiments, projects, models, etc.
        :params: query_string - graphql string to invoke.
        :params: params - variables required to invoke query string in the client.
        """
        query = gql(query_string)
        singular = self._client.execute(query, params)
        print(singular)
        return singular


    def get_collection(self, resource, query_string, params):
        """
        return a collection of objects for the requested resource.
        :params: resource - name of the resource to obtain i.e experiments, projects, models, etc.
        :params: query_string - graphql string to invoke.
        :params: params - variables required to invoke query string in the client.
        """
        query = gql(query_string)
        singular_resource =  lambda resource_name: str(resource_name.title()[0:-1]) # format resource name to match one of the existing classes
        class_name = lambda class_name: getattr(sys.modules[__name__], class_name) # convert string to class name
        collection = self._client.execute(query, params)[resource]
        class_to_init = class_name(singular_resource(resource))

        collection_objects = []
        if not resource == "jobs":
            collection_objects = [class_to_init(self._client, item, self._user_id, extract_id(item['sk'])) for item in collection]
            return collection_objects
        # jobs resource
        collection_objects = [class_to_init(self._upload_client, item, self._user_id, extract_id(item['pk'])) for item in collection]
        return collection_objects

    def projects(self):
        """
        retrieve user projects
        :params: user_id - a uuid
        """
        params = {
            "id": str(self._user_id),
        }
        return self.get_collection("projects", PROJECTS_QUERY_FRAGMENT, params)

    def experiments(self, project_id):
        """
        retreive experiments that belong to a project
        :params: project_id - a uuid
        """
        params = {
            "id": str(project_id),
        }
        return self.get_collection("experiments", EXPERIMENTS_QUERY_FRAGMENT, params)

    def experiment(self, experiment_id):
        """
        retreive experiments that belong to a project
        :params: project_id - a uuid
        """
        params = {
            "id": str(experiment_id),
        }
        return self.get_single("experiment", EXPERIMENT_QUERY_FRAGMENT, params)

    # grab user id + project id
    def project(self, project_id):
        """
        retrieve a single user project
        :params: project_id - a uuid
        """
        params = {
             "userId": str(self._user_id),
             "projectId": str(project_id)
        }
        return self.get_single("project", PROJECT_QUERY_FRAGMENT, params)

    def models(self, organization_id):
        """
        retrieve models associated with a user / organization.
        :params: project_id - a uuid
        """
        params = {
            "id": str(organization_id),
        }
        return self.get_collection("models", MODELS_QUERY_FRAGMENT, params)


    def model(self, model_id):
        """
        retrieve a single user model
        :params: project_id - a uuid
        """
        params = {
            "id": str(model_id),
        }
        return self.get_single("model", MODEL_QUERY_FRAGMENT, params)


    def jobs(self, project_id):
        """
        returns the list of jobs associated with a project
        :params: project_id - list of jobs associated with a project id
        :params: filter - condition (pending, in_prgress, data_uploaded)
        """
        params = {
            "id": str(project_id),
            "userId": self._user_id
        }
        return self.get_collection("jobs", JOBS_QUERY_FRAGMENT, params)

    def job(self, job_id):
        """
        returns a single labeling job
        :params: job_id - job uuid
        """
        params = {
            "id": str(job_id)
        }
        return self.get_single("job", EXPERIMENT_QUERY_FRAGMENT, params)


    def create_project(self, file):
        """
        create user project
        """
        with open(file, 'r') as yaml_in:
            yaml_object = yaml.safe_load(yaml_in) # yaml_object will be a list or a dict
            project_dict = yaml_object['Project']
            project_dict['userId'] = self._user_id
            now = datetime.now()
            project_dict['date'] = now.strftime("%m-%d-%Y") # We probably should not allow cli to decide DATE
            print(project_dict)
            params = project_dict
            response =  self.mutate_single(PROJECT_CREATE_FRAGMENT, params)
            new_project_created = response["createProject"]["ok"]
            return self.project(new_project_created)

        return f"Failed to open file {file}"

    def upload_class_labels(self, class_labels_file, project_id):
        """
        Precondition: create_project has been called
        This method will upload the meta.json file to the S3 storage bucket of the newly created project.
        """

        data = {}

        with open(class_labels_file) as f:
            data = json.load(f) #serialize json object to a string to be sent to server
            #upload to project_id/meta.json

            data = json.dumps(data)

            params = {
                "userId": self._user_id,
                "projectId": project_id,
                "classLabels": data
            }

            response = self.mutate_single(UPLOAD_CLASS_LABELS_MUTATION, params)


    def create_experiment(self, file):
        """
        create user experient
        """

        env = EnvYAML(file)
        project_id = env['Experiment.projectId']
        if project_id is None:
            print("no project id was set")
            return

        with open(file, 'r') as yaml_in:
            yaml_object = yaml.safe_load(yaml_in) # yaml_object will be a list or a dict
            experiment_dict = yaml_object['Experiment']
            experiment_dict['experimentId'] = "".join(str(uuid.uuid1()).split("-"))
            experiment_dict['userId'] = self._user_id
            now = datetime.now()
            experiment_dict['date'] = now.strftime("%m-%d-%Y")
            # project id from the env variable
            experiment_dict['projectId'] = project_id
            response =  self.mutate_single(EXPERIMENT_CREATE_FRAGMENT, experiment_dict)
            new_experiment_created = response["createExperiment"]["ok"]

        return self.experiment(new_experiment_created)

    # TODO:
    def create_model(self, model_path):
        """
        upload model checksum and verify there are enough models to check
        :returns: model object
        """
        return Model("", "", "", "")



    # class for pagination for class elements.



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bidirectional.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='bidirectional.proto',
  package='bidirectional',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x13\x62idirectional.proto\x12\rbidirectional\"\x1a\n\x07Message\x12\x0f\n\x07message\x18\x01 \x01(\t\"&\n\x13SeverExpStartReturn\x12\x0f\n\x07message\x18\x01 \x01(\t\"$\n\x0fStartExperiment\x12\x11\n\texp_token\x18\x01 \x01(\t\"*\n\x15GetExperimentResponse\x12\x11\n\texp_token\x18\x01 \x01(\t\"O\n\nSDKPayload\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07user_id\x18\x02 \x01(\t\x12\x12\n\nproject_id\x18\x03 \x01(\t\x12\x0e\n\x06\x65xp_id\x18\x04 \x01(\t\"\xb4\x01\n\x12\x45XPAlectioResponse\x12\x0e\n\x06status\x18\x01 \x01(\t\x12\x15\n\rexperiment_id\x18\x02 \x01(\t\x12\x12\n\nproject_id\x18\x03 \x01(\t\x12\x10\n\x08\x63ur_loop\x18\x04 \x01(\t\x12\x0f\n\x07user_id\x18\x05 \x01(\t\x12\x13\n\x0b\x62ucket_name\x18\x06 \x01(\t\x12\x0c\n\x04type\x18\x07 \x01(\t\x12\r\n\x05n_rec\x18\x08 \x01(\x02\x12\x0e\n\x06n_loop\x18\t \x01(\x02\x32\x9f\x02\n\rBidirectional\x12I\n\x11GetServerResponse\x12\x16.bidirectional.Message\x1a\x16.bidirectional.Message\"\x00(\x01\x30\x01\x12\x62\n\x1aGetStartExperimentResponse\x12\x1e.bidirectional.StartExperiment\x1a\x1e.bidirectional.StartExperiment\"\x00(\x01\x30\x01\x12_\n\x0eGetSDKResponse\x12$.bidirectional.GetExperimentResponse\x1a!.bidirectional.EXPAlectioResponse\"\x00(\x01\x30\x01\x62\x06proto3'
)




_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='bidirectional.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='bidirectional.Message.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=64,
)


_SEVEREXPSTARTRETURN = _descriptor.Descriptor(
  name='SeverExpStartReturn',
  full_name='bidirectional.SeverExpStartReturn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='bidirectional.SeverExpStartReturn.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=66,
  serialized_end=104,
)


_STARTEXPERIMENT = _descriptor.Descriptor(
  name='StartExperiment',
  full_name='bidirectional.StartExperiment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='exp_token', full_name='bidirectional.StartExperiment.exp_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=106,
  serialized_end=142,
)


_GETEXPERIMENTRESPONSE = _descriptor.Descriptor(
  name='GetExperimentResponse',
  full_name='bidirectional.GetExperimentResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='exp_token', full_name='bidirectional.GetExperimentResponse.exp_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=144,
  serialized_end=186,
)


_SDKPAYLOAD = _descriptor.Descriptor(
  name='SDKPayload',
  full_name='bidirectional.SDKPayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='bidirectional.SDKPayload.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='bidirectional.SDKPayload.user_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='bidirectional.SDKPayload.project_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exp_id', full_name='bidirectional.SDKPayload.exp_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=188,
  serialized_end=267,
)


_EXPALECTIORESPONSE = _descriptor.Descriptor(
  name='EXPAlectioResponse',
  full_name='bidirectional.EXPAlectioResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='bidirectional.EXPAlectioResponse.status', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiment_id', full_name='bidirectional.EXPAlectioResponse.experiment_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='bidirectional.EXPAlectioResponse.project_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cur_loop', full_name='bidirectional.EXPAlectioResponse.cur_loop', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='bidirectional.EXPAlectioResponse.user_id', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bucket_name', full_name='bidirectional.EXPAlectioResponse.bucket_name', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='bidirectional.EXPAlectioResponse.type', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='n_rec', full_name='bidirectional.EXPAlectioResponse.n_rec', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='n_loop', full_name='bidirectional.EXPAlectioResponse.n_loop', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=270,
  serialized_end=450,
)

DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['SeverExpStartReturn'] = _SEVEREXPSTARTRETURN
DESCRIPTOR.message_types_by_name['StartExperiment'] = _STARTEXPERIMENT
DESCRIPTOR.message_types_by_name['GetExperimentResponse'] = _GETEXPERIMENTRESPONSE
DESCRIPTOR.message_types_by_name['SDKPayload'] = _SDKPAYLOAD
DESCRIPTOR.message_types_by_name['EXPAlectioResponse'] = _EXPALECTIORESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'bidirectional_pb2'
  # @@protoc_insertion_point(class_scope:bidirectional.Message)
  })
_sym_db.RegisterMessage(Message)

SeverExpStartReturn = _reflection.GeneratedProtocolMessageType('SeverExpStartReturn', (_message.Message,), {
  'DESCRIPTOR' : _SEVEREXPSTARTRETURN,
  '__module__' : 'bidirectional_pb2'
  # @@protoc_insertion_point(class_scope:bidirectional.SeverExpStartReturn)
  })
_sym_db.RegisterMessage(SeverExpStartReturn)

StartExperiment = _reflection.GeneratedProtocolMessageType('StartExperiment', (_message.Message,), {
  'DESCRIPTOR' : _STARTEXPERIMENT,
  '__module__' : 'bidirectional_pb2'
  # @@protoc_insertion_point(class_scope:bidirectional.StartExperiment)
  })
_sym_db.RegisterMessage(StartExperiment)

GetExperimentResponse = _reflection.GeneratedProtocolMessageType('GetExperimentResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTRESPONSE,
  '__module__' : 'bidirectional_pb2'
  # @@protoc_insertion_point(class_scope:bidirectional.GetExperimentResponse)
  })
_sym_db.RegisterMessage(GetExperimentResponse)

SDKPayload = _reflection.GeneratedProtocolMessageType('SDKPayload', (_message.Message,), {
  'DESCRIPTOR' : _SDKPAYLOAD,
  '__module__' : 'bidirectional_pb2'
  # @@protoc_insertion_point(class_scope:bidirectional.SDKPayload)
  })
_sym_db.RegisterMessage(SDKPayload)

EXPAlectioResponse = _reflection.GeneratedProtocolMessageType('EXPAlectioResponse', (_message.Message,), {
  'DESCRIPTOR' : _EXPALECTIORESPONSE,
  '__module__' : 'bidirectional_pb2'
  # @@protoc_insertion_point(class_scope:bidirectional.EXPAlectioResponse)
  })
_sym_db.RegisterMessage(EXPAlectioResponse)



_BIDIRECTIONAL = _descriptor.ServiceDescriptor(
  name='Bidirectional',
  full_name='bidirectional.Bidirectional',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=453,
  serialized_end=740,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetServerResponse',
    full_name='bidirectional.Bidirectional.GetServerResponse',
    index=0,
    containing_service=None,
    input_type=_MESSAGE,
    output_type=_MESSAGE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetStartExperimentResponse',
    full_name='bidirectional.Bidirectional.GetStartExperimentResponse',
    index=1,
    containing_service=None,
    input_type=_STARTEXPERIMENT,
    output_type=_STARTEXPERIMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSDKResponse',
    full_name='bidirectional.Bidirectional.GetSDKResponse',
    index=2,
    containing_service=None,
    input_type=_GETEXPERIMENTRESPONSE,
    output_type=_EXPALECTIORESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BIDIRECTIONAL)

DESCRIPTOR.services_by_name['Bidirectional'] = _BIDIRECTIONAL

# @@protoc_insertion_point(module_scope)



"""
LabelJob Interface
"""

import json
import re
import socket
import asyncio


from gql import gql

from alectio_sdk.api.base_attribute import BaseAttribute
from alectio_sdk.api.data_upload import TextDataUpload, ImageDataUpload, NumericalDataUpload


class Job(BaseAttribute):
    def __init__(self, client, attr, user_id,  id):
        self._client = client
        self._id = id
        self._data_uploaded = attr['dataUploaded']
        self._indices = attr['indices']
        self._data_type = attr['dataType']

    def indices(self):
        """
        return the indices to upload for the job
        """
        return self._indices

    def upload_data(self, data_map):
        """
        uploads the data to be labeled for a labeling partner. primarily used in sdk to automate the job process.
        :params: data_map - data interface to be uploaded: text_file, list of image paths, or numerical file,
        example: {
            "2": path_to record,
            "5": path_to_record,
            "9": path_to_record,
        }
        keys - indices to be uploaded
        values - path to the record
        :params: data_type - text, numerical, or image
        :params: job_id - job uuid
        """
        if self._data_uploaded:
            print("data has been uploaded")
            return

        records = []
        for k,v in data_map.items():
            temp = {}
            temp['id'] = k
            temp['path'] = v
            records.append(temp)

        # grab the data type from the job attr.
        if self._data_type == "text":
            base_class = TextDataUpload(self._client)
        elif self._data_type == "image":
            base_class = ImageDataUpload(self._client)
        elif self._data_type == "numerical":
            base_class = NumericalDataUpload(self._client)
        # upload all the data asynchronously
        asyncio.get_event_loop().run_until_complete(base_class.upload_data(records, self._id))
        return None

    def __repr__(self):
        return "<Job {}>".format(self._id)



import boto3
import os
import sys
import pickle
import json
from botocore.exceptions import ClientError
import logging
from boto3.s3.transfer import TransferConfig
from threading import Thread, Lock
from io import BytesIO
import requests




class ProgressPercentage(object):
    def __init__(self, fileobj, filename, fsize=None):
        self._fileobj = fileobj
        self._filename = filename
        if not fsize:
            self._size = fileobj.getbuffer().nbytes
        else:
            self._size = fsize

        self._seen_so_far = 0
        self._lock = Lock()

    def __call__(self, bytes_amount):
        # To simplify we'll assume this is hooked up
        # to a single filename.
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\r%s %s %s %s / %s  (%.2f%%)"
                % (
                    "Uploading file ",
                    self._filename,
                    "to S3 =========>",
                    self._seen_so_far,
                    self._size,
                    percentage,
                )
            )
            sys.stdout.flush()


class S3Client:
    """Boto3 client to S3"""

    def __init__(self):
        presigned_cred = requests.post("http://prodbackend.alectio.com/experiments/getcredentials").json()
        # boto3 clients to read / write to S3 bucket
        self.client = boto3.client(
            "s3",
            aws_access_key_id=presigned_cred['KEY'],
            aws_secret_access_key=presigned_cred["SECRET"],
        )
        os.environ["AWS_ACCESS_KEY_ID"] = presigned_cred['KEY']
        os.environ["AWS_SECRET_ACCESS_KEY"] = presigned_cred['SECRET']

    def read(self, bucket_name, object_key, file_format):
        """ Read a file from the S3 bucket containing
        this experiment

        object_key: str.
           object key for this file

        file_format: str.
            format of the file {"pickle", "json"}

        """

        s3_object = self.client.get_object(Bucket=bucket_name, Key=object_key)
        body = s3_object["Body"]

        if file_format == "json":
            jstr = body.read().decode("utf-8")
            content = json.loads(jstr)
        elif file_format == "pickle":
            f = body.read()
            content = pickle.loads(f)
        elif file_format == "txt":
            content = body.read().decode(encoding="utf-8", errors="ignore")
        return content

    def write(self, obj, bucket_name, object_key, file_format):
        """Write an object to S3 bucket
        Mostly used for writing ExperimentData.pkl
        InferenceData.pkl files

        obj: dict | list | string

        bucket_name: name of the s3 bucket
        object_key: str.
            object key in the S3 bucket

        file_format: str.
            format of the file to save the object
            {pickle, json}

        """

        # convert obj to byte string
        if file_format == "pickle":
            bytestr = pickle.dumps(obj)
        elif file_format == "json":
            bytestr = json.dumps(obj)
        elif file_format == "txt":
            bytestr = b"{}".format(obj)

        # @TODO add md5 hash
        # @TODO return success or failure message
        # put in S3
        r = self.client.put_object(Bucket=bucket_name, Key=object_key, Body=bytestr,)

        return

    def multi_part_upload_with_s3(self, obj, bucket_name, object_key, file_format):
        # convert obj to byte string
        if file_format == "pickle":
            bytestr = pickle.dumps(obj)
        elif file_format == "json":
            bytestr = bytes(json.dumps(obj).encode("UTF-8"))
            # self.client.put_object(Bucket=bucket_name, Key=object_key, Body=bytestr,)
            # return
        elif file_format == "txt":
            bytestr = b"{}".format(obj)

        fileobj = BytesIO(bytestr)
        # size = sys.getsizeof(fileobj)
        # print("Size of object =" , size)

        config = TransferConfig(
            multipart_threshold=1024 * 25,
            max_concurrency=10,
            multipart_chunksize=1024 * 25,
            use_threads=True,
        )

        self.client.upload_fileobj(
            Fileobj=fileobj,
            Bucket=bucket_name,
            Key=object_key,
            Callback=ProgressPercentage(fileobj, object_key),
            Config=config,
        )

        return

    def multi_part_upload_file(self, obj, bucket_name, object_key):
        config = TransferConfig(
            multipart_threshold=1024 * 25,
            max_concurrency=10,
            multipart_chunksize=1024 * 25,
            use_threads=True,
        )

        self.client.upload_file(
            obj,
            bucket_name,
            object_key,
            Config=config,
            Callback=ProgressPercentage(obj, object_key, os.path.getsize(obj)),
        )

        return

    def upload_file(self, file_name, bucket, object_name=None):
        """Upload a file to an S3 bucket

        :param file_name: File to upload
        :param bucket: Bucket to upload to
        :param object_name: S3 object name. If not specified then file_name is used
        :return: True if file was uploaded, else False
        """

        # If S3 object_name was not specified, use file_name
        if object_name is None:
            object_name = file_name

        # Upload the file
        s3_client = boto3.client('s3')
        try:
            response = s3_client.upload_file(file_name, bucket, object_name)
        except ClientError as e:
            logging.error(e)
            return False
        return True

    def download_checkpoints(self, bucket_name, project_id, experiment_id, cur_loop, log_dir):
        checkpoints_to_download = list(range(cur_loop))

        for checkpoint in checkpoints_to_download:
            print(f"downloading file ckpt_{checkpoint}")
            self.client.download_file(
                f"{bucket_name}",
                f"{project_id}/{experiment_id}/ckpt_{checkpoint}.pth",
                f"{log_dir}/ckpt_{checkpoint}.pth",
            )



"""
mutations commands to update / create resources
"""

##### MUTATIONS #####

UPDATE_IP_PORT_MUTATION = """mutation updateProjectIp($userId: String!, $projectId: String!, $ip: String, $port: Int) {
    updateProjectIp(userId: $userId, projectId: $projectId, ip: $ip, port: $port) {
        project {
            onPremField
        }
    }
}"""


UPLOAD_QUERY_STRATEGY_MUTATION = """mutation uploadQueryStrategyMutation($queryStratData: [QueryStrategyInput], $projectId: String!, $experimentId: String!, $type: String!, $mode: String!) {
    uploadQueryStrategyMutation(queryStratData: $queryStratData, projectId:$projectId, experimentId:$experimentId, type:$type, mode:$mode) {
        ok
        message
    }
}"""

##### FILE UPLOAD MUTATIONS #####

UPLOAD_PARTNER_IMAGE_MUTATION = """mutation uploadPartnerImageMutation($files: Upload!, $records: [RecordsInput],  $jobId: String!) {
    uploadPartnerImageMutation(files: $files, records: $records, jobId: $jobId) {
        ok
    }
}"""

# TODO: support for this
UPLOAD_PARTNER_TEXT_MUTATION = """mutation uploadPartnerTextMutation($files: Upload!, $records: [RecordsInput],  $jobId: String!) {
    uploadPartnerTextMutation(files: $files, records: $records, jobId: $jobId) {
        ok
    }
}"""

# TODO: support fo this
UPLOAD_PARTNER_NUMERICAL_MUTATION = """mutation uploadPartnerNumericalMutation($files: Upload!, $records: [RecordsInput],  $jobId: String!) {
    uploadPartnerNumericalMutation(files: $files, records: $records, jobId: $jobId) {
        ok
    }
}"""

START_EXPERIMENT_MUTATION = """mutation startExperimentMutation($userId: String!, $projectId: String!, $experimentId: String!) {
    startExperimentMutation(userId: $userId, projectId: $projectId, experimentId: $experimentId) {
        ok
    }
}"""

UPLOAD_CLASS_LABELS_MUTATION = """mutation uploadClassLabelsMutation($userId: String!, $projectId: String!, $classLabels: String!) {
    uploadClassLabelsMutation(userId: $userId, projectId: $projectId, classLabels: $classLabels) {
        ok
    }
}"""

##### FILE DOWNLOAD MUTATIONS #####



"""
exceptions for alectio repository
"""

class APIKeyNotFound(Exception):     
    '''
    raise exception when api key is not set
    '''
    def __str__(self):
        message = "Alectio API Key not found"
        return message


class ModelNotFound(Exception):
    '''
    raise exceptions when the model path does not match the original checksum
    '''
    def __init__(self, val):
        self.val = val

    def __str__(self):
        message = f"Model not found for checksum: {self.val}"
        return message



class ModelExceedsChanges(Exception):
    '''
    raise exceptions when the model path does not match the original checksum
    '''
    def __init__(self, val):
        self.val = val

    def __str__(self):
        message = f"Model cannot be fully changed {self.val}"
        return message


class YAMLFieldNotFound(Exception):
    '''
    raise excpetion when the field is no found in the yaml 
    '''
    def __init__(self, val):
        self.val = val
    
    def __str__(self):
        message = f"Field not found for {self.val}"
        return message

class InvalidYAMLField(Exception):
    '''
    raise excpetion when the field field is invalid
    '''
    def __init__(self, key, val):
        self.key = key
        self.val = val
    
    def __str__(self):
        message = f"Field not found for {self.key}: {self.val}"
        return message


class InvalidYAMLFieldType(Exception):
    '''
    raise excpetion when the field field is invalid
    '''
    def __init__(self, key, type):
        self.key = key
        self.type = type
    
    def __str__(self):
        message = f"Field {self.key} must be {self.type}"
        return message


 





import sklearn.metrics
import numpy as np

class ClassificationMetrics(object):
    r"""
    Evaluation Metrics for classification

    Args:
        predictions: list,
            list of classes predicted for each record 
            
        ground_truth: list,
            list of ground_truth classes for each record
    """

    def __init__(self, predictions, ground_truth):
        self.predictions = predictions
        self.ground_truth = ground_truth
        self._compute_metrics(predictions, ground_truth)

    def _compute_metrics(self, predictions, ground_truth):
        self.confusion_matrix = sklearn.metrics.confusion_matrix(ground_truth, predictions)
        self.FP = self.confusion_matrix.sum(axis=0) - np.diag(self.confusion_matrix)
        self.FN = self.confusion_matrix.sum(axis=1) - np.diag(self.confusion_matrix)
        self.TP = self.confusion_matrix.diagonal()
        self.TN = self.confusion_matrix.sum() - (self.FP + self.FN + self.TP)

        self.precision = self.TP / (self.TP + self.FP)
        self.precision[np.isnan(self.precision)] = 0
        self.recall = self.TP / (self.TP + self.FN)
        self.recall[np.isnan(self.recall)] = 0
        self.accuracy = sklearn.metrics.accuracy_score(ground_truth, predictions)
        self.f1_score = 2 * self.precision * self.recall / (self.precision + self.recall)
        self.f1_score[np.isnan(self.f1_score)] = 0
        
        self.label_disagreement = {k: v.round(3) for k, v in enumerate(self.FP / (self.FP + self.TN))}
        self.acc_per_class = {k: v.round(3) for k, v in enumerate(self.confusion_matrix.diagonal() / self.confusion_matrix.sum(axis=1))}
        self.f1_score_per_class = {k: v for (k, v) in enumerate(self.f1_score)}
        self.f1_score = self.f1_score.mean()

        self.precision_per_class = {k: v for (k, v) in enumerate(self.precision)}
        self.precision = self.precision.mean()

        self.recall_per_class = {k: v for (k, v) in enumerate(self.recall)}
        self.recall = self.recall.mean()

        self.confusion_matrix = self.confusion_matrix.tolist()

    def get_accuracy(self):
        r"""
        Returns:
            accuracy as a float
        """
        return self.accuracy

    def get_f1_score(self):
        r"""
        Returns:
            F1 score as a float
        """
        return self.f1_score

    def get_f1_score_per_class(self):
        r"""
        Returns:
            F1 score per class as a dict
        """
        return self.f1_score_per_class

    def get_precision_per_class(self):
        r"""
        Returns:
            precision per class as a dict
        """
        return self.precision_per_class

    def get_precision(self):
        r"""
        Returns:
            precision as a float
        """
        return self.precision

    def get_recall_per_class(self):
        r"""
        Returns:
            recall per class as a dict
        """
        return self.recall_per_class

    def get_recall(self):
        r"""
        Returns:
            recall as a float
        """
        return self.recall

    def get_confusion_matrix(self):
        r"""
        Returns:
            confusion matrix as a list
        """
        return self.confusion_matrix

    def get_acc_per_class(self):
        r"""
        Returns:
            acc_per_class as a dict
        """
        return self.acc_per_class

    def get_label_disagreement(self):
        r"""
        Returns:
            label disagreement as a dict
        """
        return self.label_disagreement



"""
Project Interface
"""
import json
import re
import socket


from gql import gql

from alectio_sdk.api.base_attribute import BaseAttribute
from alectio_sdk.api.experiment import Experiment

from alectio_sdk.tools.utils import extract_id
from alectio_sdk.tools.fragments import EXPERIMENTS_QUERY_FRAGMENT
from alectio_sdk.tools.mutations import UPDATE_IP_PORT_MUTATION


class Project(BaseAttribute):

    def __init__(self, client, attr, user_id, id):
        self._client = client
        self._attr = attr # project attributes
        self._prem_info = {}
        self._name = attr['name']
        self._id = id
        self._user_id = user_id
        self._experiments = {}
        self.set_project_fields(self._attr)
        super().__init__(self._attr, self._id)


    def update_ip_port(self, ip_addr=None, port=None):
        """
        update a project's ip and port
        :params: ip_addr - ip address user intends to modify
        :params: port - port the user intends to change
        """
        params = {
            "userId": str(self._user_id),
            "projectId": str(self._id),
            "port": None,
            "ip": None
        }

        if ip_addr is None and port is None:
            raise "No fields were set."

        # check if the user inputed a valid port  0 .. 2^16
        if not port is None:
            if port > 65535 or port < 0:
                raise "Must enter a valid port number 0 < port < 65535."
            params['port'] = int(port)

        # check if the user inputed a valid ip addres x.x.x.x where x <- 0 .. 2^8-1
        if not ip_addr is None:
            try:
                socket.inet_aton(ip_addr)
                params['ip'] = ip_addr
            except socket.error:
                raise "Must enter a valid ip address x.x.x.x"

        query = gql(UPDATE_IP_PORT_MUTATION)
        updated_port_ip_query = self._client.execute(query, params)['updateProjectIp']['project']['onPremField']
        print(updated_port_ip_query)
        return None

    def experiments(self):
        """
        retreive experiments that belong to a project
        :params: project_id - a uuid
        """
        query = gql(EXPERIMENTS_QUERY_FRAGMENT)
        params = {
            "id": str(self._id),
        }
        experiments_query  = self._client.execute(query, params)['experiments']
        project_experiments = [Experiment(self._client, item, self._user_id, extract_id(item['sk'])) for item in experiments_query]
        return project_experiments


    def set_project_fields(self, attr):
        """
        set project specific fields, if the fields exists
        :params: project attributes
        """
        if 'prem_info' in attr:
            self._prem_info = json.loads(attr['prem_info'])


    def __repr__(self):
        return "<Project {}>".format(self._name)




from __future__ import print_function

import grpc
import alectio_sdk.proto.bidirectional_pb2_grpc as bidirectional_pb2_grpc
import alectio_sdk.proto.bidirectional_pb2 as bidirectional_pb2


GRPCBACKEND = "grpc.alectio.com:50051"
credentials = grpc.ssl_channel_credentials()

class Backend(object):
    def __init__(self, token):
        self.token = token
        self._done = False
        self.result_q = None
        self.response = None

    def make_start_exp_payload(self):
        yield bidirectional_pb2.StartExperiment(
            exp_token=self.token
        )
    
    def make_sdk_response_payload(self):
        yield bidirectional_pb2.GetExperimentResponse(
            exp_token=self.token
        )

    def startExperiment(self):
        # project_id = request.get_json()['project_id']
        # user_id = request.get_json()['user_id']
        # experiment_id = request.get_json()['experiment_id']
        credentials = grpc.ssl_channel_credentials()
        with grpc.secure_channel(GRPCBACKEND, credentials) as channel:
            stub = bidirectional_pb2_grpc.BidirectionalStub(channel)
            responses = stub.GetStartExperimentResponse(self.make_start_exp_payload())
            for response in responses:
                return response.exp_token
    
    def getSDKResponse(self):
        credentials = grpc.ssl_channel_credentials()
        with grpc.secure_channel(GRPCBACKEND, credentials) as channel:
            stub = bidirectional_pb2_grpc.BidirectionalStub(channel)
            responses = stub.GetSDKResponse(self.make_sdk_response_payload())
            response_obj = next(responses)
            response = {
                "status": response_obj.status,
                "experiment_id": response_obj.experiment_id,
                "project_id": response_obj.project_id,
                "cur_loop": response_obj.cur_loop,
                "user_id": response_obj.user_id,
                "bucket_name": response_obj.bucket_name,
                "type": response_obj.type,
                "n_rec": response_obj.n_rec,
                "n_loop": response_obj.n_loop
            }
            return response

def make_message(message):
    return bidirectional_pb2.StartExperiment(
        exp_token=message
    )


def generate_messages():
    messages = [
        make_message("First message"),
        make_message("Second message"),
        make_message("Third message"),
        make_message("Fourth message"),
        make_message("Fifth message"),
    ]
    for msg in messages:
        print("Hello Server Sending you the %s" % msg.exp_token)
        yield msg


def send_message(stub):
    responses = stub.GetStartExperimentResponse(generate_messages())
    for response in responses:
        print("Hello from the server received your %s" % response.exp_token)


def run():
    credentials = grpc.ssl_channel_credentials()
    with grpc.secure_channel(GRPCBACKEND, credentials) as channel:
        stub = bidirectional_pb2_grpc.BidirectionalStub(channel)
        send_message(stub)


if __name__ == '__main__':
    #run()
    backend = Backend("03af80240af4084af08320adf03040")
    backend.startExperiment()



import requests

class Downloader(object):
    def __init__(self, dataset, data_dir='data', use_current_dir=True, location="./"):
        self.dataset = dataset
        self.use_current_dir = use_current_dir
        self.location = location
    
    def download(self, dataset):
        pass
    
    def download_google_drive_file(self, id, destination):
        URL = "https://docs.google.com/uc?export=download"

        session = requests.Session()

        response = session.get(URL, params = { 'id' : id }, stream = True)
        token = self.get_confirm_token(response)

        if token:
            params = { 'id' : id, 'confirm' : token }
            response = session.get(URL, params = params, stream = True)

        self.save_response_content(response, destination)

    def get_confirm_token(self, response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value

        return None

    def save_response_content(self, response, destination):
        CHUNK_SIZE = 32768
        with open(destination, "wb") as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)
    
    def download_file(self, url, local_filename):
        local_filename = url.split('/')[-1]
        # NOTE the stream=True parameter below
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192): 
                    # If you have chunk encoded response uncomment if
                    # and set chunk_size parameter to None.
                    #if chunk: 
                    f.write(chunk)
        return local_filename



"""
Model Interface
"""
from alectio_sdk.api.base_attribute import BaseAttribute

from alectio_sdk.tools.utils import extract_id
from alectio_sdk.exceptions import ModelNotFound, ModelExceedsChanges

from pathlib import Path

import hashlib
import difflib
import re
import sys


class Model(BaseAttribute):
    def __init__(self, client, attr={}, user_id=None, id=None):
        self.client = client
        self._attr = attr
        self._id = id
        self._model_path = "" # need to get the users model path, read path from local machine 
        super().__init__(self._attr, self._id)


    def update_model(self, other): 
        """

        A user and organization has a limited amount of models to use, if a user wants to update a model, the changes
        must not exceed a certain threshold. A user must create the model object for the new model the user intends to add 
        to the list of current models. In addtion, the user needs to specify where the existing model is located based on the checksum 
        of the original model. 

        :params: other : another model to compare againse and update the existing model

        """
        # comparing the correct objects
        if not isinstance(other, Model):
            raise RuntimeError

        # check if the user has set the paths
        if len(other._model_path) == 0 or len(self._model_path) == 0:
            raise RuntimeError

        threshold = 85

        # verify the user can locate the original model
        current_checksum = self.checksum()
        original_checksum_path = self._model_path
        original_checksum = hashlib.md5(open(original_checksum_path,'rb').read()).hexdigest()
        if original_checksum != current_checksum:
            raise ModelNotFound(current_checksum)

        # read model py files as txt
        # NOTE: at the moment we do not account on filtering comments 
        with open(self._model_path) as old_model:
            old_lines = [line.rstrip().replace(" ", "") for line in old_model]

        with open(other._model_path) as updated_model:
            new_lines = [line.rstrip().replace(" ", "") for line in updated_model]
    
        # compare the contents of the files without new lines or whitespaces 
        ratio_compare = difflib.SequenceMatcher(None, old_lines, new_lines).ratio() * 100
        if ratio_compare < threshold: 
            # show the difference between the old + new
            file_difference = difflib.unified_diff(old_lines, new_lines)
            sys.stdout.writelines(file_difference)
            raise ModelExceedsChanges(ratio_compare)
        
        # need to update the new checksum, submit mutation with the following: checksum 
        other_model_path = other._model_path
        updated_file_checksum = hashlib.md5(open(other_model_path,'rb').read()).hexdigest()

        # once its gets updated, need to return the parameters back from the mutation
        # TODO: once the mutation gets created.
        return updated_file_checksum


    def checksum(self):
        """
        retreive current model checksum
        """
        return self._attr['checksum']

    @property
    def model_path(self): 
        return self._model_path

    @model_path.setter
    def model_path(self, value): 
        self._model_path = value

    
    def __repr__(self):
        return "<Model {}>".format(self._id)

   


"""
Parse YAML files and perform sanity checks on the files.
"""
import yaml

from alectio.exceptions import YAMLFieldNotFound, InvalidYAMLField, InvalidYAMLFieldType


class ParseStrategyYaml():
    """
    parse strategy yaml
    """
    def __init__(self, path):
        self._path = path
        self._required_fields = ["mode", "query_strategy", "type", "resource"]
        self._valid_query_strategies = ["random", "confidence", "margin", "entropy"]
        self._valid_modes = ["simple", "expert"]
        self._valid_intervals = ["lowest", "highest"]
        self._valid_experiment_type = ["manual_al", "regular_al"]
        self._qs_list = []
        self._experiment_mode = None
        self._experiment_type = None
        self._object = self.parse_yaml(path)
        #print(self._object)
        self.sanity_checks()


    def query_strategies_sanity(self, experiment_mode, query_strategy):
        """
        query strategies formated in the yaml
        """
        qs_names = list(query_strategy.keys())
        for qs in qs_names:
            _qs = qs
            if "_" in _qs:
                _qs = _qs.split("_")[0]
            if not _qs in self._valid_query_strategies:
                raise InvalidYAMLField("qs", qs)

        # check against the mode, each mode will have a variable amount of query strats
        if experiment_mode == "simple":
            qs_name = qs_names[0]
            query_strategy_object = query_strategy[qs_name]   
            self.simple_fields_sanity(qs_name, query_strategy_object)
            print("passed simple tests")

        else:  
            print("starting expert tests")
            self.expert_fields_sanity(qs_names, query_strategy)
            print("finished expert tests")
        return 


    '''
    query_strategy:
        confidence:
            n_rec: 100
            over: 0.5
            under: 0.6
        margin:
            n_rec: 400
            over: 0.5
            under: 0.6
        confidence_1:
            n_rec: 400
            over: 0.4
            under: 0.6
    '''
    def expert_fields_sanity(self, qs_list, qs_objects):
        """
        check expert fields for:
            - uniqueness (no QS top level key can be the same)
            - valid data type
            - completeness
        """
        #check for uniqueness of qs_object key names

        if not len(set(qs_list)) == len(qs_list):
            #duplicate names exist
            print(f"make sure all your QS's under the query_strategy key have unique names, for example confidence_1, confidence_2, etc")
            return

        for qs in qs_objects:
            obj = qs_objects[qs]
            print(qs, "->", obj)

            if "random" in qs:
                if not "n_rec" in qs_objects[qs].keys():
                    print(f"n_rec must be a key of", qs)
                    return
                if not isinstance(qs_objects[qs]["n_rec"], int):
                    print(f"n_rec must be an integer, right now it is", type(qs_objects[qs]["n_rec"]))
                    return 
            else:
                # verification of n_rec
                if not "n_rec" in qs_objects[qs].keys():
                    print(f"n_rec must be a key of", qs)
                    return
                if not isinstance(qs_objects[qs]["n_rec"], int):
                    print(f"n_rec must be an integer, right now it is", type(qs_objects[qs]["n_rec"]))
                
                #verification of over
                if not "over" in qs_objects[qs].keys():
                    print(f"You need to supply the over key for {qs}")
                    return
                if not isinstance(qs_objects[qs]["over"], float):
                    print(f"over must be an float, right now it is", type(qs_objects[qs]["over"]))
                
                #verification of under
                if not "under" in qs_objects[qs].keys():
                    print(f"You need to supply the under key for {qs}")
                    return
                if not isinstance(qs_objects[qs]["over"], float):
                    print(f"over must be an float, right now it is", type(qs_objects[qs]["over"]))

                #verification that under > over

                if not qs_objects[qs]["under"] > qs_objects[qs]["over"]:
                    print(f"under key must be greater than over key for {qs}")
                    return
        
        # all sanity checks have passed, we can add the QS objects to our qs list. 
        self._qs_list.extend(qs_objects)
        return 


    def simple_fields_sanity(self, qs, qs_object):
        """
        check fields for simple qs 
        :params: qs - query strat
        :params: qs_object - query object
        """
        qs_object_keys = list(qs_object.keys())
        query_strategy_object = qs_object
        query_strategy_object['qs'] = qs

        # generic fields that apply to random and other qs 
        if not 'n_rec' in qs_object_keys:
            raise YAMLFieldNotFound("n_rec")

        # check if it is the correct type
        n_rec = qs_object['n_rec']

        if not isinstance(n_rec, int):
            raise InvalidYAMLFieldType("n_rec", int)

        query_strategy_object['nRec'] = n_rec
        del query_strategy_object['n_rec']

        if not qs == "random":
            if not 'type' in qs_object_keys:
                raise YAMLFieldNotFound("type")

            type = qs_object['type']
            if not type in self._valid_intervals:
                raise InvalidYAMLField("type", type)

            query_strategy_object['type'] = type

        # query strat fields are good, can use.   
        self._qs_list.append(query_strategy_object)

        return 

   
    def sanity_checks(self):
        """
        perform file sanity checks to make sure the data 
        """
        yaml_object = self._object
        print(yaml_object)
        # check for outer keys
        for key in list(yaml_object.keys()):
            if key == "resource":
                continue
            elif key not in self._required_fields:
                raise YAMLFieldNotFound(key)

        # perform checks on the mode: either simple or expert 
        experiment_mode = self._object['mode']
        experiment_type = self._object['type']

        if not experiment_mode in self._valid_modes:
            raise InvalidYAMLField("type", experiment_mode)
        
        if not experiment_type in self._valid_experiment_type:
            raise InvalidYAMLField("type", experiment_type)


        self._experiment_mode = experiment_mode
        self._experiment_type = experiment_type

        qs_list = self._object['query_strategy']

        self.query_strategies_sanity(experiment_mode, qs_list)

        return 


    def parse_yaml(self, path):
        """
        parse yaml file the
        :params: path - path to yaml file.
        """
        with open(path, 'r') as stream:
            data = yaml.safe_load(stream)
        return data 


    @property
    def experiment_mode(self):
        """
        experiment mode used in the yaml,
        must be the same as when the experiment was created.
        """
        return self._experiment_mode

    @property
    def qs_list(self):
        """
        list of query strategies the user intends to use 
        """
        return self._qs_list

    @property
    def experiment_type(self):
        """
        the experiment type for the yaml
        """
        return self._experiment_type

     
# if __name__ == "__main__":
#     parser = ParseStrategyYaml("../../tests/test_files/expert_multiple_strat.yaml")


"""
query fragments
"""


##### QUERIES #####

EXPERIMENTS_QUERY_FRAGMENT = """
    query experimentsQuery($id: String!) {
        experiments(id: $id) {
            pk,
            sk,
            name,
            alType,
            nLoops,
            nRecords
        }
    }
"""

EXPERIMENT_QUERY_FRAGMENT = """
    query experimentQuery($id: String!) {
        experiment(id: $id) {
            pk,
            sk,
            name,
            alType,
            nLoops,
            nRecords
        }
    }
"""

MODELS_QUERY_FRAGMENT = """
    query modelsQuery($id: String!) {
        models(id: $id) {
            pk,
            sk,
            name,
            checksum
        }
    }
"""

MODEL_QUERY_FRAGMENT = """
    query modelQuery($id: String!) {
        model(id: $id) {
            pk,
            sk,
            name,
            checksum
        }
    }
"""

PROJECTS_QUERY_FRAGMENT = """
    query projectsQuery($id: String!) {
        projects(id: $id) {
            pk,
            sk,
            name,
            type,
            onPremField
        }
    }
"""


PROJECT_QUERY_FRAGMENT = """
    query projectQuery($userId: String!, $projectId: String!) {
        project(userId: $userId, projectId: $projectId) {
            pk,
            sk,
            name,
            type,
            onPremField
        }
    }
"""

USER_PAID_QUERY_FRAGMENT = """
    query userQuery($id: String!) {
        user(id: $id) {
            isPaid
        }
    }
"""



# TODO: refractor + reformat
EXPERIMENT_CREATE_FRAGMENT = """
    mutation CreateExperiment($userId: String!, $projectId: String!, $experimentId: String!, $name: String!, $nLoops: Int!, $nRecords: Int!, $qs: String!,
    $alType: String!, $date: String!) {
        createExperiment(userId: $userId, projectId: $projectId, experimentId: $experimentId, name: $name, nLoops: $nLoops, nRecords: $nRecords, qs: $qs, alType: $alType, date: $date) {
            ok
        }
    }
"""

PROJECT_CREATE_FRAGMENT = """
    mutation CreateProject($userId: String!, $preLabeled: String!, $alectioDataset: String!, $modelType: String!,
               $problemType: String!, $s3Bucket: String!, $premise: String!, $projectName: String!, $dataFormat: String!,
               $date: String!, $labelingType: Boolean!, $labelingCompany: Boolean!, $testLen: Int!, $trainLen: Int!, $allLabeled: Boolean!, $alectioDir: String!, $preLoadedModel: Boolean!
               $dockerUrl: String!, $ip: String!, $port: Int!){
        createProject(userId: $userId, preLabeled: $preLabeled, alectioDataset: $alectioDataset,
        modelType: $modelType, problemType: $problemType, s3Bucket: $s3Bucket, premise: $premise,
        projectName: $projectName, dataFormat: $dataFormat, date: $date, labelingType: $labelingType,
        labelingCompany: $labelingCompany, testLen: $testLen, trainLen: $trainLen, allLabeled: $allLabeled,
        alectioDir: $alectioDir, preLoadedModel: $preLoadedModel, dockerUrl: $dockerUrl, ip: $ip, port: $port) {
            ok
        }
    }
"""


JOBS_QUERY_FRAGMENT = """
    query jobsQuery($id: String!, $userId: String!) {
        jobs(id: $id, userId: $userId) {
            pk,
            sk,
            indices,
            dataUploaded,
            dataType,
        }
    }
"""



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import alectio_sdk.proto.bidirectional_pb2 as bidirectional__pb2


class BidirectionalStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetServerResponse = channel.stream_stream(
                '/bidirectional.Bidirectional/GetServerResponse',
                request_serializer=bidirectional__pb2.Message.SerializeToString,
                response_deserializer=bidirectional__pb2.Message.FromString,
                )
        self.GetStartExperimentResponse = channel.stream_stream(
                '/bidirectional.Bidirectional/GetStartExperimentResponse',
                request_serializer=bidirectional__pb2.StartExperiment.SerializeToString,
                response_deserializer=bidirectional__pb2.StartExperiment.FromString,
                )
        self.GetSDKResponse = channel.stream_stream(
                '/bidirectional.Bidirectional/GetSDKResponse',
                request_serializer=bidirectional__pb2.GetExperimentResponse.SerializeToString,
                response_deserializer=bidirectional__pb2.EXPAlectioResponse.FromString,
                )


class BidirectionalServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetServerResponse(self, request_iterator, context):
        """A Bidirectional streaming RPC.

        Accepts a stream of Message sent while a route is being traversed,
        while receiving other Message (e.g. from other users).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStartExperimentResponse(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSDKResponse(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BidirectionalServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetServerResponse': grpc.stream_stream_rpc_method_handler(
                    servicer.GetServerResponse,
                    request_deserializer=bidirectional__pb2.Message.FromString,
                    response_serializer=bidirectional__pb2.Message.SerializeToString,
            ),
            'GetStartExperimentResponse': grpc.stream_stream_rpc_method_handler(
                    servicer.GetStartExperimentResponse,
                    request_deserializer=bidirectional__pb2.StartExperiment.FromString,
                    response_serializer=bidirectional__pb2.StartExperiment.SerializeToString,
            ),
            'GetSDKResponse': grpc.stream_stream_rpc_method_handler(
                    servicer.GetSDKResponse,
                    request_deserializer=bidirectional__pb2.GetExperimentResponse.FromString,
                    response_serializer=bidirectional__pb2.EXPAlectioResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'bidirectional.Bidirectional', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Bidirectional(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetServerResponse(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/bidirectional.Bidirectional/GetServerResponse',
            bidirectional__pb2.Message.SerializeToString,
            bidirectional__pb2.Message.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetStartExperimentResponse(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/bidirectional.Bidirectional/GetStartExperimentResponse',
            bidirectional__pb2.StartExperiment.SerializeToString,
            bidirectional__pb2.StartExperiment.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSDKResponse(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/bidirectional.Bidirectional/GetSDKResponse',
            bidirectional__pb2.GetExperimentResponse.SerializeToString,
            bidirectional__pb2.EXPAlectioResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import requests
import webbrowser
import os
import json
from os import path
from uuid import uuid4
from webbrowser import open_new
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.request import urlopen, HTTPError
from urllib.parse import urlparse, parse_qs
from requests.auth import HTTPBasicAuth
from flask import jsonify

class S(BaseHTTPRequestHandler):
    def _set_response(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def do_GET(self):
        self._set_response()
        self.wfile.write("All done. You can close the window now.".encode())
        code = urlparse(self.path).query.split("=")[1]
        if code:
            try:
                data = {
                    'grant_type':'authorization_code',
                    'code':code
                    }
                response = requests.post('http://127.0.0.1:5000/oauth/token', 
                                        auth=HTTPBasicAuth(client_id, client_secret), data=data)
                print(response.json())
                with open('client_token.json', 'w') as fp:
                    json.dump(response.json(), fp)

            except:
                print("SDK Setp Failed")

            print(response)
            raise KeyboardInterrupt


    def do_POST(self):
        content_length = int(self.headers['Content-Length']) # <--- Gets the size of data
        post_data = self.rfile.read(content_length) # <--- Gets the data itself
        self._set_response()
        self.wfile.write("POST request for {}".format(self.path).encode('utf-8'))



def run(ACCESS_URI, server_class=HTTPServer, handler_class=S, port=5001):
    open_new(ACCESS_URI)
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()



def auth_call(client_id, client_secret):
    client_info = {
        'client_id': client_id,
        'client_secret':client_secret
    }
    ACCESS_URI = ('http://127.0.0.1:5000/' 
            + 'oauth/authorize?client_id=' + client_info['client_id'] + '&scope=openid+profile&response_type=code&nonce=abc')
    run(ACCESS_URI)
